/**
 * PETALS - PETALS Services Platform. Copyright (C) 2006 EBM WebSourcing
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * 
 */

package org.ow2.petals.container.lifecycle;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jbi.JBIException;
import javax.jbi.component.Bootstrap;
import javax.jbi.component.Component;
import javax.jbi.management.LifeCycleMBean;
import javax.management.AttributeChangeNotification;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanNotificationInfo;
import javax.management.Notification;
import javax.management.NotificationBroadcaster;
import javax.management.NotificationBroadcasterSupport;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.transaction.TransactionManager;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Contingency;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.communication.jndi.client.JNDIService;
import org.ow2.petals.container.ContainerService;
import org.ow2.petals.container.event.StateChangeFailedEvent;
import org.ow2.petals.container.event.StateChangedEvent;
import org.ow2.petals.container.thread.BootstrapThread;
import org.ow2.petals.container.thread.ComponentLifeCycleThread;
import org.ow2.petals.jbi.component.context.ComponentContextImpl;
import org.ow2.petals.jbi.component.context.ComponentInitialContext;
import org.ow2.petals.jbi.component.context.InstallationContextImpl;
import org.ow2.petals.jbi.descriptor.original.generated.ClassLoaderDelegationType;
import org.ow2.petals.jbi.descriptor.original.generated.ComponentType;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.descriptor.original.generated.Component.SharedLibrary;
import org.ow2.petals.jbi.management.MBeanNames;
import org.ow2.petals.jbi.management.ManagementException;
import org.ow2.petals.jbi.management.admin.AdminServiceMBean;
import org.ow2.petals.jbi.management.task.TaskContextConstants;
import org.ow2.petals.jbi.management.task.installation.InstallationContextConstants;
import org.ow2.petals.jbi.management.task.installation.InstallationUtils;
import org.ow2.petals.jbi.messaging.endpoint.ConsumerEndpoint;
import org.ow2.petals.jbi.messaging.registry.EndpointRegistry;
import org.ow2.petals.jbi.messaging.routing.Router;
import org.ow2.petals.jbi.messaging.routing.RoutingException;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.kernel.server.FractalHelper;
import org.ow2.petals.platform.classloader.ClassLoaderService;
import org.ow2.petals.platform.classloader.PetalsClassLoader;
import org.ow2.petals.platform.repository.RepositoryService;
import org.ow2.petals.platform.systemstate.SystemStateService;
import org.ow2.petals.util.LoggingUtil;

/**
 * @author alouis
 * @author Olivier Fabre - EBM Websourcing
 * @author wjoseph - eBMWebsourcing
 * 
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.container.lifecycle.InstallerMBean.class))
public class Installer implements InstallerMBean, NotificationBroadcaster,
        ComponentContextCommunication {

    /**
     * Bootstrap classloader id prefix.
     */
    public static final String BS_PREFIX = "BS";

    /**
     * Component classloader id prefix.
     */
    public static final String COMP_PREFIX = "COMP";

    public static final String INSTALLED = "Installed";

    public static final String INSTALLING = "Installing";

    public static final String UNINSTALLED = "Uninstalled";

    public static final String UNINSTALLING = "Uninstalling";

    public static final String UNKNOWN = "Unknown";

    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * JBI-Management :: Admin Service
     */
    @Requires(name = "admin", signature = org.ow2.petals.jbi.management.admin.AdminServiceMBean.class)
    protected AdminServiceMBean adminService;

    /**
     * Platform :: Loader Service
     */
    @Requires(name = "classloader", signature = org.ow2.petals.platform.classloader.ClassLoaderService.class)
    protected ClassLoaderService loadSrv;

    /**
     * JBI-Messaging :: Router Service
     */
    @Requires(name = "router", signature = org.ow2.petals.jbi.messaging.routing.Router.class)
    protected Router router;

    /**
     * Platform :: Repository Service
     */
    @Requires(name = "repository", signature = org.ow2.petals.platform.repository.RepositoryService.class)
    protected RepositoryService repositorySrv;

    /**
     * JBI-Messaging :: EndpointRegistry
     */
    @Requires(name = "endpoint", signature = org.ow2.petals.jbi.messaging.registry.EndpointRegistry.class)
    protected EndpointRegistry EndpointRegistry;

    /**
     * Communication :: JNDIService
     */
    @Requires(name = "jndi", signature = org.ow2.petals.communication.jndi.client.JNDIService.class)
    protected JNDIService jndiService;

    /**
     * Container :: Container Service
     */
    @Requires(name = "container", signature = org.ow2.petals.container.ContainerService.class)
    protected ContainerService container;

    /**
     * JBI-Management :: TransactionManager Service
     */
    @Requires(name = "transactionManager", signature = javax.transaction.TransactionManager.class, contingency = Contingency.OPTIONAL)
    protected TransactionManager transactionManager;

    /**
     * Platform component :: SystemState Service
     */
    @Requires(name = "systemstate", signature = org.ow2.petals.platform.systemstate.SystemStateService.class)
    protected SystemStateService recoverySrv;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * The Fractal component
     */
    @Service
    private org.objectweb.fractal.api.Component fractalComponent;

    /**
     * The container configuration
     */
    private ContainerConfiguration containerConfiguration;

    protected HashMap<String, Object> context;

    /**
     * The ObjectName of this installer
     */
    protected ObjectName mBeanName;

    /**
     * A sequence number used by state change notification
     */
    protected long sequenceNumber;

    protected Jbi descriptor;

    protected String installationRootPath;

    protected String workingRootPath;

    protected ConsumerEndpoint address;

    protected InitialContext initialContext;

    private InstallationContextImpl installContext;

    private PetalsClassLoader bootstrapClassLoader;

    protected BootstrapThread bootstrapThread;

    // /////////////////
    // Notifications support
    // /////////////////
    protected Object activitySynchronizer;

    protected String state;

    NotificationBroadcasterSupport notifSupport = new NotificationBroadcasterSupport();

    protected ComponentContextImpl componentContext;

    protected org.ow2.petals.jbi.descriptor.original.generated.Component componentDescription;

    protected ObjectName componentJmxName;

    protected URI installationRoot;

    protected Bootstrap jbiBootstrap;

    protected LoggingUtil log;

    /**
     * Create the Installer that will manage installation / uninstallation of
     * the JBI component. The related bootstrap is created and initialize here.
     */
    public Installer() {
        this.state = "";
    }

    /**
     * Initialize the installer.
     * 
     * @param ctxt
     *      the installation context
     */
    public void init(Map<String, Object> ctxt) throws PetalsException, JBIException {

        this.context = (HashMap<String, Object>) ctxt;

        this.state = UNINSTALLED;

        this.activitySynchronizer = new Object();

        this.descriptor = (Jbi) this.context.get(TaskContextConstants.DESCRIPTOR);

        this.installationRoot = (URI) this.context.get(TaskContextConstants.INSTALL_ROOT);

        this.componentDescription = this.descriptor.getComponent();

        this.mBeanName = this.adminService.getMBeanNames().createInstallerMBeanName(
                this.componentDescription.getIdentification().getName());

        String componentName = InstallationUtils.getComponentNameFromDescriptor(this.descriptor);

        this.installationRootPath = new File(this.installationRoot).getAbsolutePath();

        this.workingRootPath = this.repositorySrv.getComponentWorkRoot(componentName)
                .getAbsolutePath();

        this.address = new ConsumerEndpoint(componentName, containerConfiguration.getName(),
                containerConfiguration.getSubdomainName());

        this.initialContext = this.createComponentInitialContext();

        // 1. create the component context
        this.componentContext = this.createComponentContext();

        // 2. create the bootstrap classloader
        this.bootstrapClassLoader = this.createBootstrapClassloader();

        // 3. create the bootstrap object
        this.jbiBootstrap = this.loadBootstrap(this.bootstrapClassLoader);

        // 5. create installation context
        this.installContext = new InstallationContextImpl(this.componentDescription,
                this.componentContext, true);

        // 6. init the threadWorker that execute actions on the Bootstrap
        this.bootstrapThread = this.createBootstrapThread(this.bootstrapClassLoader,
                this.jbiBootstrap);

        // 7. init bootstrap
        this.initBootstrap(true);

    }

    public void initBootstrap(boolean install) throws JBIException {
        this.installContext.setInstall(install);
        try {
            this.bootstrapThread.init(this.installContext);
        } catch (JBIException e) {
            this.bootstrapThread.cleanUp();
            throw e;
        }

    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.NotificationBroadcaster#addNotificationListener(javax
     * .management.NotificationListener, javax.management.NotificationFilter,
     * java.lang.Object)
     */
    public void addNotificationListener(NotificationListener arg0, NotificationFilter arg1,
            Object arg2) throws IllegalArgumentException {
        this.notifSupport.addNotificationListener(arg0, arg1, arg2);
    }

    /**
     * Get the unique component name of this component, as assigned by the
     * identification section of this component's installation descriptor.
     * 
     * @return the component name; must be non-null and non-empty.
     */
    public String getComponentName() {
        return this.componentContext.getComponentName();
    }

    // ////////////////////////////////////////////
    // ElementListener implementation
    // ////////////////////////////////////////////
    /**
     * @see javax.jbi.management.LifeCycleMBean#getCurrentState()
     */
    public String getCurrentState() {
        synchronized (this.state) {
            if (!(LifeCycleMBean.SHUTDOWN.equals(this.state)
                    || LifeCycleMBean.STARTED.equals(this.state) || LifeCycleMBean.STOPPED
                    .equals(this.state))) {
                return LifeCycleMBean.UNKNOWN;
            }
        }
        return this.state;
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.management.InstallerMBean#getInstallerConfigurationMBean()
     */
    public ObjectName getInstallerConfigurationMBean() throws JBIException {
        log.call();

        if (this.jbiBootstrap == null) {
            throw new JBIException("Bootstrap should not be null");
        }

        return this.bootstrapThread.getExtensionMBeanName();
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.management.InstallerMBean#getInstallRoot()
     */
    public String getInstallRoot() {
        return this.componentContext.getInstallRoot();
    }

    /**
     * 
     * @return the installer ObjectName
     */
    public ObjectName getMBeanName() {
        return this.mBeanName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.NotificationBroadcaster#getNotificationInfo()
     */
    public MBeanNotificationInfo[] getNotificationInfo() {
        String[] types = new String[] { AttributeChangeNotification.ATTRIBUTE_CHANGE };

        String name = AttributeChangeNotification.class.getName();

        String description = "Notification about a change of the lifeCycle state or a state change failure.";

        MBeanNotificationInfo info = new MBeanNotificationInfo(types, name, description);

        return new MBeanNotificationInfo[] { info };
    }

    /**
     * Install the jbi component. Call the Bootstrap onInstall() and cleanup()
     * methods, register the jbiComponent object, and return its jmxName.
     * 
     * @see javax.jbi.management.InstallerMBean#install()
     * 
     * @throws JBIException
     *      the component is already installed
     */
    public ObjectName install() throws JBIException {
        this.log.start();
        if (UNINSTALLED.equals(this.state)) {
            synchronized (this.activitySynchronizer) {
                try {
                    this.setState(INSTALLING);
                    this.doInstall();
                    this.setState(INSTALLED);
                } catch (Throwable re) {
                    this.setState(UNKNOWN);

                    String msg = "An exception occured during the installation.";

                    JBIException e = new JBIException(msg, re);

                    this.log.error(msg, e);

                    this.stateChangeFailed(new StateChangeFailedEvent(this, e));

                    throw e;
                }
            }
        } else {
            String msg = "The Object can not be installed in this state: " + this.state;

            JBIException e = new JBIException(msg);

            this.log.error(msg, e);

            this.stateChangeFailed(new StateChangeFailedEvent(this, e));

            throw e;
        }
        this.log.end();

        return this.componentJmxName;
    }

    /**
     * Test the installation flag
     * 
     * @return true if the component is already installed
     */
    public boolean isInstalled() {
        boolean result = false;
        if (INSTALLED.equals(this.state)) {
            result = true;
        }
        return result;
    }

    /*
     * recoverySrv (non-Javadoc)
     * 
     * @see
     * javax.management.NotificationBroadcaster#removeNotificationListener(javax
     * .management.NotificationListener)
     */
    public void removeNotificationListener(NotificationListener arg0)
            throws ListenerNotFoundException {
        this.notifSupport.removeNotificationListener(arg0);
    }

    /**
     * Set the <code>State</code> reference of the current state. An
     * <code>LifeCycleStateEventAbstract</code> is thrown.
     * 
     * @param state
     * @throws JBIException
     */
    public void setState(String state) throws JBIException {
        synchronized (this.state) {
            this.stateChanged(new StateChangedEvent(this, this.state, state));
            this.state = state;
        }

        try {
            this.recoverySrv.updateComponentInstallationState(this.getComponentName(), state);
        } catch (Exception e) {
            String msg = "Component state can't be persisted for recovery";
            this.log.error(msg, e);
            throw new JBIException(msg, e);
        }
    }

    /**
     * <i>stateChanged</i> is called when the state of a lifecycle element
     * changes. A JMX-notification is thrown.
     * 
     * @see
     *  petals.management.element.ElementStateListener#stateChanged(java.util
     *  .EventObject)
     */
    public void stateChanged(StateChangedEvent event) {
        this.log.call();
        synchronized (this.state) {
            Notification n = new AttributeChangeNotification(this.mBeanName, this.sequenceNumber++,
                    System.currentTimeMillis(), "State of '" + this.mBeanName.toString()
                            + "' changed : " + event.getOldState() + " -> " + event.getNewState(),
                    "state", "String", event.getOldState(), event.getNewState());

            this.notifSupport.sendNotification(n);
        }
    }

    /**
     * <i>stateChangeFailed</i>> is called when the change of state of a
     * lifecycle element fails. A JMX-notification throws the relative
     * JBIException, .
     * 
     * @see
     *  petals.management.element.ElementStateListener#stateChanged(java.util
     *  .EventObject)
     */
    public void stateChangeFailed(StateChangeFailedEvent event) {
        this.log.call();
        synchronized (this.state) {
            Notification n = new AttributeChangeNotification(this.mBeanName, this.sequenceNumber++,
                    System.currentTimeMillis(), "State change of '" + this.mBeanName.toString()
                            + "' failed : " + event.getException().getMessage(), "exception",
                    "String", null, event.getException().toString());

            this.notifSupport.sendNotification(n);
        }
    }

    /**
     * Uninstall the JBI component. Call the Bootstrap onUninstall() and
     * cleanup() methods, unregister the jbiComponent object,
     * 
     * @throws JBIException
     *      the component is already uninstalled
     */
    public void uninstall() throws JBIException {
        this.log.start();

        ComponentLifeCycle componentLifeCycle = this.getComponentLifeCycle();
        if (componentLifeCycle.getServiceUnitLifeCycles().size() > 0) {
            String msg = "Can not uninstall a component which holds deployed Service Unit(s)";
            this.log.error(msg);
            throw new JBIException(msg);
        }

        String compState = componentLifeCycle.getCurrentState();
        if (INSTALLED.equals(this.state)
                && (LifeCycleMBean.SHUTDOWN.equals(compState) || LifeCycleMBean.UNKNOWN
                        .equals(compState))) {

            synchronized (this.activitySynchronizer) {
                try {
                    this.setState(UNINSTALLING);
                    this.doUninstall();
                    this.setState(UNINSTALLED);
                } catch (Throwable re) {
                    this.setState(UNKNOWN);
                    String msg = "An exception occured during the uninstallation of the component "
                            + this.componentContext.getComponentName();
                    JBIException e = new JBIException(msg, re);
                    this.log.error(msg, e);
                    this.stateChangeFailed(new StateChangeFailedEvent(this, e));
                    throw e;
                }
            }
        } else {
            String msg = "The component '" + this.componentContext.getComponentName()
                    + "' can not be uninstalled in the Installer state: " + this.state
                    + " and Component state: " + compState;
            JBIException e = new JBIException(msg);
            this.log.error(msg, e);
            this.stateChangeFailed(new StateChangeFailedEvent(this, e));
            throw e;
        }

        this.log.end();
    }

    /**
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     *  getAddress()
     */
    public ConsumerEndpoint getAddress() {
        return this.address;
    }

    /**
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     *  getEndpointRegistry()
     */
    public EndpointRegistry getEndpointRegistry() {
        return this.EndpointRegistry;
    }

    /**
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     *  getInitialContext()
     */
    public InitialContext getInitialContext() {
        return this.initialContext;
    }

    /**
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     *  getInstallationRoot()
     */
    public String getInstallationRoot() {
        return this.installationRootPath;
    }

    /**
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     *  getJBIDescriptor()
     */
    public Jbi getJBIDescriptor() {
        return this.descriptor;
    }

    /**
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     *  getAdminService()
     */
    public AdminServiceMBean getAdminService() {
        return this.adminService;
    }

    /**
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     *  getRouterService()
     */
    public Router getRouterService() {
        return this.router;
    }

    /**
     * @see org.ow2.petals.container.lifecycle.ComponentContextCommunication#
     *  getTransactionManagerService()
     */
    public TransactionManager getTransactionManagerService() {
        return this.transactionManager;
    }

    /**
     * Get the workspace root
     * 
     * @return the workspace root
     */
    public String getWorkspaceRoot() {
        return this.workingRootPath;
    }

    /**
     * Get the logger
     * 
     * @return the logger
     */
    public Logger getLogger() {
        return this.logger;
    }

    /**
     * Shutdown all pending actions in the Installer and clean attributes (the
     * threadworker is shutdown)
     * 
     * @throws JBIException
     */
    public void shutdown() throws JBIException {
        this.context = null;
        this.descriptor = null;
        this.installationRoot = null;
        this.componentDescription = null;
        this.mBeanName = null;
        this.installationRootPath = null;
        this.workingRootPath = null;
        this.address = null;
        this.initialContext = null;
        this.componentContext = null;
        this.bootstrapClassLoader = null;
        this.jbiBootstrap = null;
        this.installContext = null;

        this.bootstrapThread.shutdownThread();
        this.bootstrapThread = null;
    }

    /**
     * Create the classLoader for Bootstrap
     * 
     * @return classLoader for Bootstrap
     */
    protected PetalsClassLoader createBootstrapClassloader() {
        this.log.call();

        List<String> classpath = this.componentDescription.getBootstrapClassPath().getPathElement();
        // trim the list of class path elements
        for (int i = 0; i < classpath.size(); i++) {
            classpath.get(i).trim();
        }

        boolean parentFirst = !ClassLoaderDelegationType.SELF_FIRST
                .equals(this.componentDescription.getBootstrapClassLoaderDelegation());

        return this.createClassLoader(BS_PREFIX, classpath, parentFirst, null);
    }

    /**
     * Create a classloader with the specified parameters.
     * 
     * @param prefix
     *      bootstrap or component prefix
     * @param classpath
     *      classpath of the bootstrap/component
     * @param parentFirst
     *      delegation mode
     * @param slNameList
     *      list of the name of the required sharedLibraries
     * @return a classloader
     */
    protected PetalsClassLoader createClassLoader(String prefix, List<String> classpath,
            boolean parentFirst, List<String> slNameList) {
        this.log.start();

        // prepare bootstrap and component classloader
        List<URL> urls = new ArrayList<URL>();

        try {
            urls.add(this.installationRoot.normalize().toURL());
        } catch (MalformedURLException e) {
            // simple do nothing - JBI Package has responsibility in
            // checking the archive contents against informations
            // contained in JBI Descriptor first.
            String msg = "Unexpected error. " + e.getMessage();
            this.log.info(msg, e);
        }
        URL[] baseUrls = urls.toArray(new URL[] {});

        PetalsClassLoader classLoader = null;
        try {
            classLoader = (PetalsClassLoader) this.loadSrv.createComponentClassLoader(prefix
                    + this.componentDescription.getIdentification().getName(), baseUrls, classpath,
                    parentFirst, slNameList);
        } catch (PetalsException pe) {
            // just log when some error occur
            this.log.error("Error creating bootstrap class loader. " + pe.getMessage(), pe);
        }

        this.log.end();
        return classLoader;
    }

    /**
     * Create the classloader for the component. Be careful, we use the
     * classpath set in the InstallationContext. (during the installation phase,
     * the Bootstrap can modify the component's classpath by setting it in the
     * InstallationContext object.)
     * 
     * @return classloader for the component
     */
    @SuppressWarnings("unchecked")
    protected PetalsClassLoader createComponentClassloader() {
        this.log.start();

        List<String> classpath = this.installContext.getClassPathElements();

        // Delegation model
        boolean parentFirst = !ClassLoaderDelegationType.SELF_FIRST
                .equals(this.componentDescription.getComponentClassLoaderDelegation());

        // Shared libraries used by this component
        List<SharedLibrary> slList = this.componentDescription.getSharedLibraryList();
        List slNameList = new ArrayList();
        if (slList != null) {
            for (SharedLibrary sl : slList) {
                slNameList.add(sl.getContent().trim());
            }
        }

        PetalsClassLoader componentClassloader = this.createClassLoader(COMP_PREFIX, classpath,
                parentFirst, slNameList);

        this.log.end();
        return componentClassloader;
    }

    /**
     * Do the bootstrap installation.
     * 
     * @throws JBIException
     * @throws PetalsException
     */
    protected void doInstall() throws JBIException, PetalsException {
        this.log.start();

        try {
            this.bootstrapThread.onInstall();
        } finally {
            this.bootstrapThread.cleanUp();
        }

        // create the component life cycle
        this.componentJmxName = this.createComponentLifeCycleMBean();

        this.log.end();
    }

    /**
     * Create the thread that will be used to process component installation
     * steps. The thread-classloader is set with the given paremeter.
     * 
     * @param bootstrap
     *      classloader
     * @param bootstrap
     *      instance
     * @return the BootstrapThread
     */
    protected BootstrapThread createBootstrapThread(PetalsClassLoader classloader,
            Bootstrap bootstrap) {
        BootstrapThread it = new BootstrapThread(this.getComponentName(), bootstrap,
                this.containerConfiguration.getTaskTimeout());
        it.setContextClassLoader(classloader);
        it.start();
        return it;
    }

    /**
     * Create the thread that will be used to process component life cycle
     * steps. The thread-classloader is set with the given paremeter.
     * 
     * @param classloader
     * @return the BootstrapThread
     */
    protected ComponentLifeCycleThread createComponentLifeCycleThread(Component component,
            PetalsClassLoader classloader) {
        ComponentLifeCycleThread clt = new ComponentLifeCycleThread(component,
                containerConfiguration.getTaskTimeout());
        clt.setContextClassLoader(classloader);
        clt.start();
        return clt;
    }

    /**
     * Register the installer engine/binding in the life cycle manager service
     * 
     * @return
     * @throws JBIException
     * @throws PetalsException
     */
    protected ObjectName createComponentLifeCycleMBean() throws JBIException, PetalsException {
        this.log.start();

        ObjectName objectName = null;

        // 1. create component classloader
        PetalsClassLoader cl = this.createComponentClassloader();

        // 2. create the component object
        Component jbiComponent = this.loadComponent(cl);

        // 3. set the jbiComponent reference to the contextImpl
        this.componentContext.setComponent(jbiComponent);

        // 4. create the componentLifeCycle thread
        ComponentLifeCycleThread lifeCycleThread = this.createComponentLifeCycleThread(
                jbiComponent, cl);

        // 5. create the Mbean (as binding or engine)
        MBeanNames mbeanNames = this.adminService.getMBeanNames();
        if (ComponentType.BINDING_COMPONENT.equals(this.componentDescription.getType())) {
            objectName = mbeanNames.createBindingComponentMBeanName(this.componentDescription
                    .getIdentification().getName());
        } else if (ComponentType.SERVICE_ENGINE.equals(this.componentDescription.getType())) {
            objectName = mbeanNames.createEngineComponentMBeanName(this.componentDescription
                    .getIdentification().getName());
        }

        try {
            // 6. Creation of the lifecycle
            org.objectweb.fractal.api.Component lifecycle = this.container
                    .createComponentLifeCycle(this.componentDescription);

            // 7. Start the new lifecycle
            FractalHelper.startComponent(lifecycle);

            // Initialization of the lifecycle
            ((ComponentLifeCycleMBean) lifecycle.getFcInterface("service")).init(
                    this.componentDescription, this.componentContext, jbiComponent, objectName,
                    lifeCycleThread, this.log, this.context);

            // 8. inform the router about the component installation
            this.router.addComponent(this.componentContext);

            // 9. register engine or binding
            if (ComponentType.BINDING_COMPONENT.equals(this.componentDescription.getType())) {
                this.adminService.registerBindingComponent((ComponentLifeCycle) lifecycle
                        .getFcInterface("/content"));
            } else if (ComponentType.SERVICE_ENGINE.equals(this.componentDescription.getType())) {
                this.adminService.registerEngineComponent((ComponentLifeCycle) lifecycle
                        .getFcInterface("/content"));
            }
        } catch (RoutingException e) {
            throw new JBIException(e);
        } catch (NoSuchInterfaceException e) {
            throw new JBIException(e);
        } catch (IllegalLifeCycleException e) {
            throw new JBIException(e);
        }

        this.log.end();
        return objectName;
    }

    /**
     * Unregister the installer engine/binding from the life cycle manager
     * service
     * 
     * @throws ManagementException
     */
    protected void unregisterComponentLifeCycleMBean() throws ManagementException {
        if (ComponentType.BINDING_COMPONENT.equals(this.componentDescription.getType())) {
            this.adminService.unregisterBindingComponent(this.componentJmxName);
        } else if (ComponentType.SERVICE_ENGINE.equals(this.componentDescription.getType())) {
            this.adminService.unregisterEngineComponent(this.componentJmxName);
        }
    }

    /**
     * Do the bootstrap uninstallation
     * 
     * @throws JBIException
     */
    protected void doUninstall() throws JBIException {

        this.log.start();
        if (this.jbiBootstrap == null) {
            throw new JBIException("Bootstrap should not be null");
        }

        // Ensure that the DeliveryChannel of the component is closed
        // (if the component is in an unknown state, for instance)
        if (this.componentContext.getDeliveryChannelImpl() != null
                && this.componentContext.getDeliveryChannelImpl().isOpened()) {
            this.log.debug("close the DeliveryChannel");
            this.componentContext.getDeliveryChannelImpl().close();
        }

        // call uninstall on bootstrap
        try {
            this.bootstrapThread.onUninstall();
        } finally {
            // clean the bootstrap thread
            this.bootstrapThread.cleanUp();
        }

        // unregister the component life-cycle mbean
        this.unregisterComponentLifeCycleMBean();

        // unregister the optional extension mbean
        if (this.componentContext.getComponent().getLifeCycle() != null) {
            ObjectName extName = this.componentContext.getComponent().getLifeCycle()
                    .getExtensionMBeanName();
            try {
                this.adminService.getMBeanServer().unregisterMBean(extName);
            } catch (Exception e) {
                // nothing
            }
        }

        // inform the router about the component uninstallation
        try {
            this.router.removeComponent(this.componentContext);
        } catch (RoutingException e) {
            throw new JBIException(e);
        }

        // unload component and bootstrap class loaders
        String componentName = this.componentDescription.getIdentification().getName();

        this.loadSrv.deleteClassLoader(Installer.COMP_PREFIX + componentName);
        this.loadSrv.deleteClassLoader(Installer.BS_PREFIX + componentName);

        // remove the fractal component
        try {
            this.container.removeComponentLifeCycle(this.componentDescription);
        } catch (PetalsException e) {
            throw new JBIException(e);
        }

        // clean the component context
        this.componentContext.clear();

        this.log.end();
    }

    /**
     * Get the component life cycle
     * 
     * @return the component life cycle
     */
    protected ComponentLifeCycle getComponentLifeCycle() {
        String compName = this.componentDescription.getIdentification().getName();
        return this.adminService.getComponentLifeCycleByName(compName);
    }

    /**
     * load the bootstrap, with its own classloader
     * 
     * @return
     */
    protected Bootstrap loadBootstrap(PetalsClassLoader classloader) throws JBIException {
        this.log.call();
        Class<?> bootstrapClass;
        Bootstrap jbiBootstrapTmp = null;
        try {
            bootstrapClass = classloader.loadClass(this.componentDescription
                    .getBootstrapClassName().trim());

            if (bootstrapClass == null) {
                throw new JBIException("Could not find bootstrap class: "
                        + this.componentDescription.getBootstrapClassName());
            }
            jbiBootstrapTmp = (Bootstrap) bootstrapClass.newInstance();

        } catch (ClassNotFoundException e) {
            throw new JBIException("Class not found: "
                    + this.componentDescription.getBootstrapClassName(), e);
        } catch (InstantiationException e) {
            throw new JBIException("Could not instantiate : "
                    + this.componentDescription.getBootstrapClassName(), e);
        } catch (IllegalAccessException e) {
            throw new JBIException("Illegal access on: "
                    + this.componentDescription.getBootstrapClassName(), e);
        }
        return jbiBootstrapTmp;
    }

    /**
     * load the component, with its own classloader
     * 
     * @return the component
     */
    protected javax.jbi.component.Component loadComponent(PetalsClassLoader classloader)
            throws JBIException {
        this.log.call();

        Class<?> componentClass;
        Component jbiComponent = null;

        try {
            componentClass = classloader.loadClass(this.componentDescription
                    .getComponentClassName().getContent().trim());

            if (componentClass == null) {
                throw new JBIException("Could not find component class: "
                        + this.componentDescription.getComponentClassName());
            }

            jbiComponent = (javax.jbi.component.Component) componentClass.newInstance();

        } catch (ClassNotFoundException e) {
            throw new JBIException("Class not found: "
                    + this.componentDescription.getComponentClassName(), e);
        } catch (InstantiationException e) {
            throw new JBIException("Could not instantiate : "
                    + this.componentDescription.getComponentClassName(), e);
        } catch (IllegalAccessException e) {
            throw new JBIException("Illegal access on: "
                    + this.componentDescription.getComponentClassName(), e);
        } catch (SecurityException e) {
            throw new JBIException("Security exception for java.util.logging.Logger", e);
        }

        return jbiComponent;
    }

    /**
     * Initialization of the petals component
     * 
     * @throws PetalsException
     * @throws JBIException
     * 
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws PetalsException, JBIException {
        this.log = new LoggingUtil(this.logger, this.logger.getName().substring(
                this.logger.getName().lastIndexOf(".") + 1));
        this.log.start();

        containerConfiguration = configurationService.getContainerConfiguration();

        this.log.end();
    }

    /**
     * Stop of the petals component
     * 
     */
    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() {
        this.log.call();
        this.logger = null;
    }

    /**
     * Create and return the component context
     * 
     * @return the component context
     * @throws PetalsException
     */
    private ComponentContextImpl createComponentContext() throws PetalsException {

        ComponentContextImpl componentContextImpl;
        try {
            componentContextImpl = new ComponentContextImpl(
                    (ComponentContextCommunication) this.fractalComponent
                            .getFcInterface("/content"));
        } catch (NoSuchInterfaceException e) {
            throw new PetalsException(e);
        }

        /*
         * Fill context
         */
        this.context.put(InstallationContextConstants.COMPONENT_CONTEXT, componentContextImpl);

        return componentContextImpl;
    }

    /**
     * The initial JNDI context for all component. If this initial context is
     * null, it will be created
     * 
     * @return the initial JNDI context for all component. If this initial
     *  context is null, it will be created
     * @throws PetalsException
     *      if {@link InitialContext} can't be created
     */
    private InitialContext createComponentInitialContext() throws PetalsException {
        ComponentInitialContext componentInitialContext = null;
        try {
            componentInitialContext = new ComponentInitialContext(this.jndiService
                    .getUsersContext());
        } catch (NamingException e) {
            String msg = "Error creating InitialContext for components";
            this.log.error(msg, e);
            throw new PetalsException(msg, e);
        }
        return componentInitialContext;
    }

}
