/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 Fossil E-Commerce,
 * http://www.fossilec.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: ComponentContextImpl.java 9231 2008-09-25 14:06:20Z rnaudin $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.component.context;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.Set;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.Component;
import javax.jbi.management.MBeanNames;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.naming.InitialContext;
import javax.xml.namespace.QName;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.Handler;
import org.objectweb.util.monolog.api.MonologFactory;
import org.objectweb.util.monolog.api.TopicalLogger;
import org.ow2.petals.container.lifecycle.ComponentContextCommunication;
import org.ow2.petals.container.lifecycle.ComponentLifeCycle;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.messaging.endpoint.AbstractEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.ConsumerEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.InternalEndpoint;
import org.ow2.petals.jbi.messaging.exchange.DeliveryChannelImpl;
import org.ow2.petals.jbi.messaging.routing.Router;
import org.ow2.petals.jbi.security.SecurityContext;
import org.ow2.petals.util.LoggingUtil;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;

import static org.ow2.petals.container.ContainerService.COMPONENT_LOGGER_PREFIX;

/**
 * @version $Revision: 9231 $ $Date: 2006-08-04 12:11:16 +0200 (ven., 04 août
 *  2006) $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 * @author wjoseph - eBMWebsourcing
 */
public class ComponentContextImpl implements javax.jbi.component.ComponentContext {

    protected DeliveryChannelImpl deliveryChannel;

    /**
     * references of registered external serviceEndpoints
     */
    protected Set<ServiceEndpoint> externalEndpoints;

    /**
     * references of registered internal serviceEndpoints
     */
    protected Set<AbstractEndpoint> internalEndpoints;

    protected Component jbiComponent;

    /**
     * The Monolog loggerFactory
     */
    private MonologFactory loggerFactory;

    /**
     * The component context communication (can be either an installer or a
     * component life cycle)
     */
    private ComponentContextCommunication componentContextCommunication;

    protected LoggingUtil logger;

    /**
     * Component loggers
     */
    private Map<String, String> componentLoggerRBNames;

    private SecurityContext securityContext;

    /**
     * Constructor of the ComponentContextImpl class
     * 
     * @param fractalComponent
     *      the fractal representation of installer
     * @throws IllegalArgumentException
     *      Fractal Error: Impossible to get the service interface
     */
    public ComponentContextImpl(ComponentContextCommunication componentContextCommunication) {
        super();

        this.componentContextCommunication = componentContextCommunication;
        org.objectweb.util.monolog.api.Logger log = this.getLogger();
        this.logger = new LoggingUtil(log, log.getName().substring(
                log.getName().lastIndexOf(".") + 1));
        this.componentLoggerRBNames = new HashMap<String, String>();
        this.internalEndpoints = new HashSet<AbstractEndpoint>();
        this.externalEndpoints = new HashSet<ServiceEndpoint>();
    }

    /**
     * @see javax.jbi.component.ComponentContext#activateEndpoint(javax.xml.
     *  namespace.QName, java.lang.String)
     */
    public ServiceEndpoint activateEndpoint(QName serviceName, String endpointName)
            throws JBIException {
        this.logger.start();
        if (serviceName == null) {
            throw new IllegalArgumentException("serviceName must be non-null");
        }
        if (endpointName == null) {
            throw new IllegalArgumentException("endpointName must be non-null");
        }
        InternalEndpoint se = this.componentContextCommunication.getEndpointRegistry()
                .activateEndpoint(serviceName, endpointName,
                        this.componentContextCommunication.getAddress());
        this.internalEndpoints.add(se);
        this.logger.end();
        return se;
    }

    /**
     * @see javax.jbi.component.ComponentContext#deactivateEndpoint(javax.jbi.
     *  servicedesc.ServiceEndpoint)
     */
    public void deactivateEndpoint(ServiceEndpoint endpoint) throws JBIException {
        this.logger.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("endpoint must be non-null");
        }
        this.componentContextCommunication.getEndpointRegistry().deactivateEndpoint(endpoint);
        this.internalEndpoints.remove(endpoint);
        this.logger.end();
    }

    /**
     * deregister all endpoints (internal and external)
     * 
     */
    public void deregisterAllEndpoints() throws JBIException {
        this.logger.start();

        for (AbstractEndpoint se : this.internalEndpoints) {
            this.componentContextCommunication.getEndpointRegistry().deactivateEndpoint(se);
        }
        this.internalEndpoints.clear();

        for (ServiceEndpoint se : this.externalEndpoints) {
            this.componentContextCommunication.getEndpointRegistry().deregisterExternalEndpoint(se);
        }
        this.externalEndpoints.clear();

        this.logger.end();
    }

    /**
     * @see
     *  javax.jbi.component.ComponentContext#deregisterExternalEndpoint(javax
     *  .jbi.servicedesc.ServiceEndpoint)
     */
    public void deregisterExternalEndpoint(ServiceEndpoint externalEndpoint) throws JBIException {
        this.logger.start();
        this.componentContextCommunication.getEndpointRegistry().deregisterExternalEndpoint(
                externalEndpoint);
        this.externalEndpoints.remove(externalEndpoint);
        this.logger.end();
    }

    public ConsumerEndpoint getAddress() {
        this.logger.call();
        return this.componentContextCommunication.getAddress();
    }

    public Component getComponent() {
        return this.jbiComponent;
    }

    /**
     * @throws JBIException
     * @see javax.jbi.component.ComponentContext#getComponentName()
     */
    public String getComponentName() {
        return this.componentContextCommunication.getJBIDescriptor().getComponent()
                .getIdentification().getName();
    }

    /**
     * @see javax.jbi.component.ComponentContext#getDeliveryChannel()
     */
    public DeliveryChannel getDeliveryChannel() throws MessagingException {
        this.logger.start();

        if (this.deliveryChannel == null) {
            this.deliveryChannel = this.createDeliveryChannel();
        } else if (!this.deliveryChannel.isOpened()) {
            this.deliveryChannel = this.createDeliveryChannel();
        } else {
            // Do not throw an exception to be TCK and other containers
            // compliant
            // throw new MessagingException("Delivery Channel already open");
        }

        this.logger.end();
        return this.deliveryChannel;
    }

    /**
     * Return the actual deliveryChannel. Can be null.
     * 
     * @return
     */
    public DeliveryChannelImpl getDeliveryChannelImpl() {
        return this.deliveryChannel;
    }

    /**
     * @throws JBIException
     * @see
     *  javax.jbi.component.ComponentContext#getEndpoint(javax.xml.namespace.
     *  QName, java.lang.String)
     */
    public ServiceEndpoint getEndpoint(QName service, String name) {
        this.logger.start();
        if (service == null) {
            throw new IllegalArgumentException("service must not be null");
        }
        if (name == null) {
            throw new IllegalArgumentException("name must not be null");
        }
        ServiceEndpoint result = this.componentContextCommunication.getEndpointRegistry()
                .getEndpoint(service, name, false);
        this.logger.end();
        return result;
    }

    /**
     * @see
     *  javax.jbi.component.ComponentContext#getEndpointDescriptor(javax.jbi.
     *  servicedesc.ServiceEndpoint)
     */
    public Document getEndpointDescriptor(ServiceEndpoint endpoint) throws JBIException {
        this.logger.start();
        Document result = null;
        if (endpoint == null) {
            throw new IllegalArgumentException("endpoint must not be null");
        }
        if (this.getEndpoint(endpoint.getServiceName(), endpoint.getEndpointName()) != null) {
            /*
             * Checks if the endpoint exist and if it is activated
             */
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) endpoint;
            if (abstractEndpoint.getEndpointRegistry() == null) {
                abstractEndpoint.setEndpointRegistry(this.componentContextCommunication
                        .getEndpointRegistry());
            }
            result = abstractEndpoint.getDescription();
        }
        this.logger.end();
        return result;
    }

    /**
     * @see
     *  javax.jbi.component.ComponentContext#getEndpoints(javax.xml.namespace
     *  .QName)
     */
    public ServiceEndpoint[] getEndpoints(QName interfaceName) {
        this.logger.start();
        ServiceEndpoint[] result = this.componentContextCommunication.getEndpointRegistry()
                .getInternalEndpointsForInterface(interfaceName);
        this.logger.end();
        return result;
    }

    /**
     * @see
     *  javax.jbi.component.ComponentContext#getEndpointsForService(javax.xml
     *  .namespace.QName)
     */
    public ServiceEndpoint[] getEndpointsForService(QName serviceName) {
        this.logger.start();
        ServiceEndpoint[] result = this.componentContextCommunication.getEndpointRegistry()
                .getInternalEndpointsForService(serviceName);
        this.logger.end();
        return result;
    }

    /**
     * @see javax.jbi.component.ComponentContext#getExternalEndpoints(javax.xml.
     *  namespace.QName)
     */
    public ServiceEndpoint[] getExternalEndpoints(QName interfaceName) {
        this.logger.start();
        ServiceEndpoint[] result = this.componentContextCommunication.getEndpointRegistry()
                .getExternalEndpointsForInterface(interfaceName);
        this.logger.end();
        return result;
    }

    /**
     * @throws JBIException
     * @see javax.jbi.component.ComponentContext#getExternalEndpointsForService(
     *  javax.xml.namespace.QName)
     */
    public ServiceEndpoint[] getExternalEndpointsForService(QName serviceName) {
        this.logger.start();
        ServiceEndpoint[] result = this.componentContextCommunication.getEndpointRegistry()
                .getExternalEndpointsForService(serviceName);
        this.logger.end();
        return result;
    }

    /**
     * @see javax.jbi.component.ComponentContext#getInstallRoot()
     */
    public String getInstallRoot() {
        this.logger.call();
        String result = this.componentContextCommunication.getInstallationRoot();
        return result;
    }

    /**
     * @see javax.jbi.component.ComponentContext#getLogger(java.lang.String,
     *  java.lang.String)
     */
    public Logger getLogger(String suffix, String resourceBundleName)
            throws MissingResourceException, JBIException {
        this.logger.start();
        String loggerName = this.getComponentName();

        // append the suffix to the logger Name
        if (suffix == null) {
            throw new IllegalArgumentException("suffix must not be null");
        } else {
            if (suffix.length() > 0) {
                loggerName += "." + suffix;
            }
        }

        // get the corresponding logger
        Logger jdkLogger = this.getComponentLogger(loggerName, resourceBundleName);

        this.logger.end();
        return jdkLogger;
    }

    /**
     * @throws JBIException
     * @see javax.jbi.component.ComponentContext#getMBeanNames()
     */
    public MBeanNames getMBeanNames() {
        this.logger.call();
        return this.componentContextCommunication.getAdminService().getMBeanNames();
    }

    /**
     * @throws JBIException
     * @see javax.jbi.component.ComponentContext#getMBeanServer()
     */
    public MBeanServer getMBeanServer() {
        this.logger.call();
        return this.componentContextCommunication.getAdminService().getMBeanServer();
    }

    /**
     * @throws JBIException
     * @see javax.jbi.component.ComponentContext#getNamingContext()
     */
    public InitialContext getNamingContext() {
        this.logger.call();
        return this.componentContextCommunication.getInitialContext();
    }

    /**
     * @see javax.jbi.component.ComponentContext#getTransactionManager()
     */
    public Object getTransactionManager() {
        this.logger.call();
        return componentContextCommunication.getTransactionManagerService();
    }

    /**
     * @throws JBIException
     * @see javax.jbi.component.ComponentContext#getWorkspaceRoot()
     */
    public String getWorkspaceRoot() {
        this.logger.call();
        return componentContextCommunication.getWorkspaceRoot();
    }

    /**
     * @see
     *  javax.jbi.component.ComponentContext#registerExternalEndpoint(javax.jbi
     *  .servicedesc.ServiceEndpoint)
     */
    public void registerExternalEndpoint(ServiceEndpoint externalEndpoint) throws JBIException {
        this.logger.start();
        if (externalEndpoint == null) {
            throw new IllegalArgumentException("externalEndpoint must not be null");
        }
        this.componentContextCommunication.getEndpointRegistry().registerExternalEndpoint(
                externalEndpoint);
        this.externalEndpoints.add(externalEndpoint);
        this.logger.end();
    }

    // //////////////////////////////////////////////////////
    // Implementation specific
    // //////////////////////////////////////////////////////

    /**
     * TODO resolveEndpointReference : refactor
     * 
     * @see
     *  javax.jbi.component.ComponentContext#resolveEndpointReference(org.w3c
     *  .dom.DocumentFragment)
     */
    public ServiceEndpoint resolveEndpointReference(DocumentFragment epr) {
        this.logger.start();
        if (epr == null) {
            throw new IllegalArgumentException("epr must not be null");
        }
        ServiceEndpoint result = null;
        // Search for all engines components if one can resolve the EPR.
        Map<ObjectName, ComponentLifeCycle> engines = this.componentContextCommunication
                .getAdminService().getEngineCompoLifeCycles();

        synchronized (engines) {
            for (Iterator<ComponentLifeCycle> iter = engines.values().iterator(); iter.hasNext()
                    && result == null;) {
                ComponentLifeCycle lifeCycle = iter.next();

                result = lifeCycle.getComponent().resolveEndpointReference(epr);
            }
        }

        if (result == null) {
            // Search for all binding components if one can resolve the EPR.
            Map<ObjectName, ComponentLifeCycle> bindings = this.componentContextCommunication
                    .getAdminService().getBindingCompoLifeCycles();

            synchronized (bindings) {
                for (Iterator<ComponentLifeCycle> iter = bindings.values().iterator(); iter
                        .hasNext()
                        && result == null;) {
                    ComponentLifeCycle lifeCycle = iter.next();

                    result = lifeCycle.getComponent().resolveEndpointReference(epr);
                }
            }
        }
        this.logger.end();
        return result;

    }

    public void setComponent(Component component) {
        this.jbiComponent = component;
    }

    public void setDeliveryChannel(DeliveryChannelImpl deliveryChannel) {
        this.deliveryChannel = deliveryChannel;
    }

    protected DeliveryChannelImpl createDeliveryChannel() {
        return new DeliveryChannelImpl(this);
    }

    public Jbi getJbi() {
        return this.componentContextCommunication.getJBIDescriptor();
    }

    /**
     * The method creates or retrieve a logger for the component.
     * 
     * @param loggerName
     * @param resourceBundleName
     * 
     * @return the jdk form logger
     * 
     * @throws JBIException
     */
    protected Logger getComponentLogger(String loggerName, String resourceBundleName)
            throws JBIException {
        this.logger.start();

        /*
         * As specified in the JBI specification it's forbidden to call a same
         * logger with two different Resource bundle names
         */
        String rbName = componentLoggerRBNames.get(loggerName);
        if (rbName != null && !rbName.equals(resourceBundleName)) {
            throw new JBIException("Logger already exists for suffix " + loggerName + " but with "
                    + rbName + " as resource bundle instead of the expected " + resourceBundleName);
        }

        Logger jdkLogger = null;

        // get a reference on the Monolog LoggerFactory
        if (this.loggerFactory == null) {
            this.loggerFactory = Monolog.getMonologFactory();
        }

        // get a new logger or retrieve it
        TopicalLogger newLogger = (TopicalLogger) this.loggerFactory.getLogger(
                COMPONENT_LOGGER_PREFIX + "." + loggerName, resourceBundleName);

        newLogger.setAdditivity(false);

        // if the logger already exits, return it directly
        if (newLogger.getHandler().length == 0) {

            // retrieve the generic console handler for components
            Handler genericHandler = this.loggerFactory.getHandler("componentsConsole");

            // create the console handler for the new logger
            Handler componentHandler = this.loggerFactory.createHandler(this.getComponentName(),
                    "console");

            // copy all the attributes from the generic handler to the new
            // handler
            for (String attributeName : genericHandler.getAttributeNames()) {
                if (!attributeName.equals("handlertype")) {
                    componentHandler.setAttribute(attributeName, genericHandler
                            .getAttribute(attributeName));
                }
            }

            // customize the pattern and activate the handler
            componentHandler.setAttribute(Handler.PATTERN_ATTRIBUTE, "["
                    + COMPONENT_LOGGER_PREFIX.toLowerCase() + "." + loggerName.toLowerCase()
                    + "]-%l %d %m%n");
            componentHandler.setAttribute("activation", this.loggerFactory);

            // add the handler to the new logger
            try {
                newLogger.addHandler(componentHandler);
                newLogger.addHandler(this.loggerFactory.getHandler("petalsLF"));
            } catch (Exception e) {
                throw new JBIException(e);
            }

            // Add loggerName to registered rbNames for loggers
            componentLoggerRBNames.put(loggerName, resourceBundleName);
        }

        // cast the logger to fit the JDK Logger class
        jdkLogger = (Logger) newLogger;

        this.logger.end();
        return jdkLogger;
    }

    public Router getRouter() {
        return this.componentContextCommunication.getRouterService();
    }

    public org.objectweb.util.monolog.api.Logger getLogger() {
        return this.componentContextCommunication.getLogger();
    }

    /**
     * Clear the component Context, remove objects dependent to the component
     * life cycle scope. To be used when to component is uninstalling.
     */
    public void clear() {
        this.deliveryChannel = null;
        this.jbiComponent = null;
    }

    /**
     * @return the securityContext
     */
    public SecurityContext getSecurityContext() {
        return securityContext;
    }

    /**
     * @param securityContext
     *      the securityContext to set
     */
    public void setSecurityContext(SecurityContext securityContext) {
        this.securityContext = securityContext;
    }

}
