/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: InstallationServiceMBean.java 154 2006-03-27 15:30:10Z alouis $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.installation;

import javax.jbi.JBIException;

import org.ow2.petals.PetalsLifeCycle;
import org.ow2.petals.container.lifecycle.Installer;
import org.ow2.petals.container.lifecycle.LifeCycleAbstract;
import org.ow2.petals.kernel.api.server.PetalsException;

/**
 * This custom interface add methods used by reinstallation tasks of the
 * SystemStateService service.
 * 
 * @author alouis
 * 
 */
public interface InstallationServiceMBean extends
    javax.jbi.management.InstallationServiceMBean, PetalsLifeCycle {

    /**
     * Get the list of components that use the shared library
     * 
     * @param slName
     *            the name of the shared library
     * @return the list of components that uses this shared library
     * @throws PetalsException
     */
    String[] retrieveComponentsForInstalledSharedlibrary(String slName)
        throws PetalsException;

    /**
     * Get the list of all installed shared libraries
     * 
     * @return the list of all installed shared libraries
     * @throws PetalsException
     */
    String[] retrieveInstalledSharedLibraries() throws PetalsException;

    /**
     * Try to start all components. If a component can't be started, the
     * starting process isn't aborted. It returns the list of all started
     * components.
     * 
     * @return an array of succesfuly started component names.
     */
    String[] startAllComponents();

    /**
     * Try to stop all components. If a component can't be stoped, the stoping
     * process isn't aborted. It returns the list of all stoped components.
     * 
     * @return an array of succesfuly stoped component names.
     */
    String[] stopAllComponents();

    /**
     * Try to shutdown all components. If a component can't be shutdown, the
     * shutdowning process isn't aborted. It returns the list of all shutdown
     * components.
     * 
     * @return an array of succesfuly shutdown component names.
     */
    String[] shutdownAllComponents();
    
    /**
     * Try to uninstall all components. If a component can't be uninstalled, the
     * uninstalling process isn't aborted. It returns the list of all uninstalled
     * components.
     * 
     * @return an array of succesfuly uninstalled component names.
     */
    String[] uninstallAllComponents();

    /**
     * Start the component. This puts the component into the
     * {@link LifeCycleAbstract#STARTED} state.
     * 
     * @param componentName
     *            name of the component to be started; must be non-null and
     *            non-empty
     * @return true if the component is correctly started false otherwise
     * @throws java.lang.Exception
     *             if there is no such component
     * @throws java.lang.Exception
     *             if the component fails to start
     */
    boolean startComponent(String componentName) throws JBIException;

    /**
     * Stop the component. This puts the component into the
     * {@link LifeCycleAbstract#STOPPED} state.
     * 
     * @param componentName
     *            name of the component to be stopped; must be non-null and
     *            non-empty
     * @return true if the component is correctly stopped false otherwise
     * @throws java.lang.Exception
     *             if there is no such component
     * @throws java.lang.Exception
     *             if the component fails to Stop
     */
    boolean stopComponent(String componentName) throws JBIException;

    /**
     * Shutdown the component. This puts the component into the
     * {@link LifeCycleAbstract#SHUTDOWN} state.
     * 
     * @param componentName
     *            name of the component to be shutdown; must be non-null and
     *            non-empty
     * @return true if the component is correctly shutdown false otherwise
     * @throws java.lang.Exception
     *             if there is no such component
     * @throws java.lang.Exception
     *             if the component fails to shutdown
     */
    boolean shutdownComponent(String componentName) throws JBIException;

    /**
     * Uninstall the component. This puts the component into the
     * {@link Installer#UNINSTALLED} state.
     * 
     * @param componentName
     *            name of the component to be uninstalled; must be non-null and
     *            non-empty
     * @return true if the component is correctly uninstalled false otherwise
     * @throws java.lang.Exception
     *             if there is no such component
     * @throws java.lang.Exception
     *             if the component fails to uninstall
     */
    boolean uninstallComponent(String componentName) throws JBIException;

    /**
     * Try to uninstall all components. If the boolean is set to true, the
     * uninstallation process is forced as this : if the component is in the
     * STARTED state it's stopped and shutdown, if the component is in the
     * STOPPED state it's shutdown. If the installer is in the INSTALLED state,
     * it's uninstalled. If a component can't be uninstalled, the uninstalling
     * process isn't aborted. It returns the list of all uninstalled components.
     * 
     * @return an array of succesfuly uninstalled component names.
     */
    String[] unloadAllComponentInstallers(boolean force);

    /**
     * Try to uninstall all shared libraries. If a shared library can't be
     * uninstalled, the uninstalling process isn't aborted. It returns the list
     * of all uninstalled shared libraries.
     * 
     * @return an array of succesfuly uninstalled shared library names.
     */
    String[] uninstallAllSharedLibraries();

    /**
     * Try to unload installer of a component. If the component is in the
     * STARTED state it's stopped and shutdown, if the component is in the
     * STOPPED state it's shutdown. If the installer is in the INSTALLED state,
     * it's uninstalled.
     * 
     * @param compName
     *            the component unique name
     * @return true if the component installer is correctly unloaded, false
     *         otherwise
     */
    boolean forceUnloadComponentInstaller(String compName);

}
