/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: DeployAllSUTask.java 154 6 oct. 06 ofabre $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.task.deployment.deploy;

import java.io.InputStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.ow2.petals.container.lifecycle.ServiceAssemblyLifeCycle;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.jbi.descriptor.original.generated.ServiceUnit;
import org.ow2.petals.jbi.management.ManagementException;
import org.ow2.petals.jbi.management.task.TaskContextConstants;
import org.ow2.petals.jbi.management.task.deployment.DeploymentContextConstants;
import org.ow2.petals.jbi.management.task.deployment.DeploymentTaskFactory;
import org.ow2.petals.jbi.management.task.deployment.DeploymentUtils;
import org.ow2.petals.jbi.management.util.XMLResult;
import org.ow2.petals.processor.Task;
import org.ow2.petals.processor.TaskProcessor;
import org.ow2.petals.util.LoggingUtil;

/**
 * This task creates service assembly lifecycle element and for each Service
 * Unit included into the service assembly, executes the deployment process of a
 * service unit <br/> Inputs : {@link Jbi}, sa install root {@link URI},
 * {@link XMLResult}, SUInstallRoots {@link Map}, SU task processors
 * {@link List} <br/> Outputs : {@link ServiceAssemblyLifeCycle} <br/>
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class DeployAllSUTask implements Task {

    protected DeploymentTaskFactory deploymentTaskFactory;

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    public DeployAllSUTask(DeploymentTaskFactory deploymentTaskFactory, LoggingUtil log) {
        super();
        this.deploymentTaskFactory = deploymentTaskFactory;
        this.log = log;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.processor.Task#execute(java.util.HashMap)
     */
    @SuppressWarnings("unchecked")
    public void execute(HashMap<String, Object> context) throws Exception {
        URI installationRoot = (URI) context.get(TaskContextConstants.INSTALL_ROOT);

        Jbi descriptor = (Jbi) context.get(TaskContextConstants.DESCRIPTOR);

        XMLResult xmlResult = (XMLResult) context.get(DeploymentContextConstants.XML_RESULT);

        Map<String, URI> suInstallRoots = (Map<String, URI>) context
                .get(DeploymentContextConstants.SU_INSTALL_ROOTS);

        List<TaskProcessor> suTaskProcessors = new ArrayList<TaskProcessor>();
        context.put(DeploymentContextConstants.SU_TASK_PROCESSORS, suTaskProcessors);

        ServiceAssemblyLifeCycle saLifeCycle = (ServiceAssemblyLifeCycle) context
                .get(DeploymentContextConstants.SA_LIFECYCLE);

        /*
         * For each Service Unit included into the service assembly : execute
         * the deployment process of a service unit
         */
        deploySUS(installationRoot, saLifeCycle, descriptor, suInstallRoots, xmlResult,
                suTaskProcessors);
    }

    /**
     * * Deploy all Service Units to the JBI container
     * 
     * @param installationRoot
     *            the root {@link URI} of the service assembly
     * @param saLifeCycle
     *            the {@link ServiceAssemblyLifeCycle} of the service assembly
     * @param descriptor
     *            the service assembly jbi descriptor
     * @param suInstallRoots
     *            a map of all SU install roots
     * @param xmlResult
     *            xml task result
     * @throws ManagementException
     *             if an sus deployment fails caused by a severe error of the
     *             container
     * 
     */
    protected void deploySUS(URI installationRoot, ServiceAssemblyLifeCycle saLifeCycle,
            Jbi descriptor, Map<String, URI> suInstallRoots, XMLResult xmlResult,
            List<TaskProcessor> suTaskProcessors) throws Exception {
        List<ServiceUnit> serviceUnits = descriptor.getServiceAssembly().getServiceUnit();
        if (serviceUnits != null) {

            StringBuffer componentOutput = new StringBuffer();
            for (ServiceUnit serviceUnit : serviceUnits) {
                // Create Deployment context
                HashMap<String, Object> suDeploymentContext = createSUDeploymentContext(
                        saLifeCycle, suInstallRoots, serviceUnit);

                // Process SU deployment
                deploySU(suDeploymentContext, suTaskProcessors);

                String xmlReturn = (String) suDeploymentContext
                        .get(DeploymentContextConstants.COMPONENT_XML_RESULT);
                componentOutput.append(xmlReturn);
            }
            xmlResult.addComponentTaskResult(componentOutput);
        }
    }

    private HashMap<String, Object> createSUDeploymentContext(ServiceAssemblyLifeCycle saLifeCycle,
            Map<String, URI> suInstallRoots, ServiceUnit serviceUnit) {
        HashMap<String, Object> suDeploymentContext = new HashMap<String, Object>();
        suDeploymentContext.put(DeploymentContextConstants.SA_LIFECYCLE, saLifeCycle);
        suDeploymentContext.put(DeploymentContextConstants.SU_DESCRIPTOR, serviceUnit);
        suDeploymentContext.put(DeploymentContextConstants.SU_INSTALL_ROOT, suInstallRoots
                .get(DeploymentUtils.getServiceUnitName(serviceUnit)));

        return suDeploymentContext;
    }

    /**
     * Check if a component task result is valid against an xml schema
     * 
     * @param componentTask
     *            xml of the component task
     * @param xsd
     *            an {@link InputStream} of the xsd file
     * @return true if valid, false if not valid
     */
    protected boolean isValidComponentTask(String componentTask, InputStream xsd) {
        return XMLResult.isValidComponentTask(componentTask, xsd);
    }

    /**
     * Process the deployment of a given SU
     * 
     * @param suDeploymentContext
     *            the deployment context of the given SU
     * @param suTaskProcessors
     *            a storage that handles all previously executed SU deployment
     *            processor
     * @throws Exception
     *             if an error occured during deployment process
     */
    private void deploySU(HashMap<String, Object> suDeploymentContext,
            List<TaskProcessor> suTaskProcessors) throws Exception {

        /*
         * Create processor and add tasks
         */
        TaskProcessor processor = new TaskProcessor(suDeploymentContext, log);
        processor.addTask(deploymentTaskFactory.getDeploySUToComponentTask());
        processor.addTask(deploymentTaskFactory.getCreateAndRegisterSULifeCycleTask());

        /*
         * Launch processor
         */
        Throwable result = processor.process();
        if (result != null) {
            // Unprocess all previously processed SuDeploymentProcessors
            for (TaskProcessor taskProcessor : suTaskProcessors) {
                if (!(taskProcessor == processor)) {
                    taskProcessor.unprocess();
                }
            }
            throw new Exception("Error occured during the deployment of a Service Unit", result);
        }

        suTaskProcessors.add(processor);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.processor.Task#undo(java.util.HashMap)
     */
    public void undo(HashMap<String, Object> context) {
        List<?> suTaskProcessors = (List<?>) context
                .get(DeploymentContextConstants.SU_TASK_PROCESSORS);
        log.debug("Revert DeployAllSUTask");
        if (suTaskProcessors != null) {
            log.debug("Unprocessing deployment of " + suTaskProcessors.size() + " SU");
            for (Object processor : suTaskProcessors) {
                ((TaskProcessor) processor).unprocess();
            }
        }
    }

}
