/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.management.task.installation.uninstall.sl;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.jbi.management.LifeCycleMBean;

import org.ow2.petals.container.lifecycle.ComponentLifeCycle;
import org.ow2.petals.jbi.descriptor.original.generated.Component;
import org.ow2.petals.jbi.descriptor.original.generated.Component.SharedLibrary;
import org.ow2.petals.jbi.management.admin.AdminServiceMBean;
import org.ow2.petals.jbi.management.task.TaskContextConstants;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.processor.Task;
import org.ow2.petals.util.LoggingUtil;

/**
 * This task checks that all components that makes use of the given SL are in
 * the shutdown state <br/> Inputs : shared lib name
 * 
 * @author ofabre
 * 
 */
public class CheckComponentShutdownStateTask implements Task {

    private AdminServiceMBean admin;

    private LoggingUtil log;

    public CheckComponentShutdownStateTask(AdminServiceMBean admin, LoggingUtil log) {
        super();
        this.admin = admin;
        this.log = log;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.processor.Task#execute(java.util.HashMap)
     */
    public void execute(HashMap<String, Object> context) throws Exception {
        String slName = (String) context.get(TaskContextConstants.ENTITY_NAME);

        if (!checkComponentShutdownStateForSL(slName)) {
            throw new PetalsException("Shared library " + slName + " can't be uninstalled because "
                    + "at least one component isn't shut down");
        }
    }

    /**
     * Check that all components that makes use of the given sl are in the
     * shutdown state.
     * 
     * @param slName
     *            the SL identifier
     * @return true if all components that makes use of the given SL are in the
     *         shutdown state, false otherwise
     */
    protected boolean checkComponentShutdownStateForSL(String slName) {
        boolean result = true;

        /*
         * Check for bindings
         */
        Collection<ComponentLifeCycle> bindings = admin.getBindingCompoLifeCycles().values();
        for (Iterator<ComponentLifeCycle> iter = bindings.iterator(); iter.hasNext() && result;) {
            ComponentLifeCycle cycle = (ComponentLifeCycle) iter.next();

            // Retrieve the list of referenced SL
            Component description = cycle.getComponentDescription();
            List<SharedLibrary> sls = description.getSharedLibraryList();

            if (sls != null) {
                for (SharedLibrary sl : sls) {
                    // If a used component is not in the SHUTDOWN state, SL
                    // can't be uninstalled
                    if (slName.equals(sl.getContent())
                            && !(LifeCycleMBean.SHUTDOWN.equals(cycle.getCurrentState()))) {
                        log.error("The shared lib can't be uninstalled because the following "
                                + "component isn't in the shutdown state : "
                                + description.getIdentification().getName());
                        result = false;
                    }
                }
            }
        }

        /*
         * Check for engines
         */
        Collection<ComponentLifeCycle> engines = admin.getEngineCompoLifeCycles().values();
        for (Iterator<ComponentLifeCycle> iter = engines.iterator(); iter.hasNext() && result;) {
            ComponentLifeCycle cycle = (ComponentLifeCycle) iter.next();

            // Retrieve the list of referenced SL
            Component description = cycle.getComponentDescription();
            List<SharedLibrary> sls = description.getSharedLibraryList();

            if (sls != null) {
                for (SharedLibrary sl : sls) {
                    // If a used component is not in the SHUTDOWN state, SL
                    // can't be uninstalled
                    if (slName.equals(sl.getContent())
                            && !(LifeCycleMBean.SHUTDOWN.equals(cycle.getCurrentState()))) {
                        log.error("The shared lib " + slName + " can't be uninstalled because"
                                + " the following component isn't in the shutdown state : "
                                + description.getIdentification().getName());
                        result = false;
                    }
                }
            }
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.processor.Task#undo(java.util.HashMap)
     */
    public void undo(HashMap<String, Object> context) throws Exception {
        //nothing to do
    }

}
