/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id: EndpointRegistry.java 11:27:27 ddesjardins $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.jbi.messaging.registry;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.communication.jmx.JMXService;
import org.ow2.petals.communication.jmx.RemoteJMXServer;
import org.ow2.petals.communication.jndi.client.JNDIService;
import org.ow2.petals.jbi.management.admin.AdminServiceMBean;
import org.ow2.petals.jbi.messaging.endpoint.AbstractEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.ConsumerEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.InternalEndpoint;
import org.ow2.petals.jbi.messaging.endpoint.LinkedEndpoint;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.util.LoggingUtil;
import org.ow2.petals.util.ParameterCheckHelper;
import org.ow2.petals.util.WSDLUtil;
import org.ow2.petals.util.XMLUtil;
import org.w3c.dom.Document;

import static org.ow2.petals.kernel.api.Constants.UNRESOLVED_INTERFACE;

/**
 * This class is used to manage everything that deals with endpoint in the
 * distributed environment
 * 
 * @author ddesjardins - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.jbi.messaging.registry.EndpointRegistry.class))
public class EndpointRegistryImpl implements EndpointRegistry {

    /**
     * Logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * Manager of the distributed JMX server
     */
    @Requires(name = "jmx", signature = org.ow2.petals.communication.jmx.JMXService.class)
    protected JMXService jmxService;

    /**
     * JNDI service fractal component
     */
    @Requires(name = "jndi", signature = org.ow2.petals.communication.jndi.client.JNDIService.class)
    protected JNDIService jndiService;

    /**
     * Admin service fractal component
     */
    @Requires(name = "admin", signature = org.ow2.petals.jbi.management.admin.AdminServiceMBean.class)
    protected AdminServiceMBean adminService;

    /**
     * Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    protected ConfigurationService configurationService;

    /**
     * JNDI Registry instance
     */
    protected JNDIRegistry registry;

    /**
     * The local container name.
     */
    private String containerName;

    /**
     * @see EndpointRegistry#activateEndpoint(QName, String, ConsumerEndpoint)
     */
    public InternalEndpoint activateEndpoint(QName serviceName, String endpointName,
            ConsumerEndpoint address) throws JBIException {
        this.log.start();

        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        if (endpointName == null) {
            throw new IllegalArgumentException("Endpoint name must be non null");
        }
        if (address == null) {
            throw new IllegalArgumentException("Consumer endpoint must be non null");
        }
        InternalEndpoint se = new InternalEndpoint(serviceName, endpointName, address
                .getComponentName(), address.getContainerName(), address.getSubDomainName(), this);
        this.registry.registerInternalEndpoint(se);

        this.log.end();
        return se;
    }

    /**
     * @see EndpointRegistry#createConnection(QName, QName, String)
     */
    public void createConnection(QName consInterface, QName provService, String provEndpoint)
            throws JBIException {
        this.log.start();

        if (consInterface == null) {
            throw new IllegalArgumentException("Interface name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consInterface, provService,
                provEndpoint, this);
        this.registry.registerConnection(linkedEndpoint);

        this.log.end();
    }

    /**
     * @see EndpointRegistry#createConnection(QName, String, QName, String)
     */
    public void createConnection(QName consService, String consEndpoint, QName provService,
            String provEndpoint) throws JBIException {
        this.log.start();

        if (consService == null) {
            throw new IllegalArgumentException("Consumer service name must be non null");
        }
        if (consEndpoint == null) {
            throw new IllegalArgumentException("Consumer endpoint name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consService, consEndpoint, provService,
                provEndpoint, this);
        this.registry.registerConnection(linkedEndpoint);

        this.log.end();
    }

    /**
     * @see EndpointRegistry#deactivateEndpoint(ServiceEndpoint)
     */
    public void deactivateEndpoint(ServiceEndpoint endpoint) throws JBIException {
        this.log.start();

        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        this.registry.unregisterInternalEndpoint((AbstractEndpoint) endpoint);

        this.log.end();
    }

    /**
     * @see EndpointRegistry#deleteConnection(QName, QName, String)
     */
    public void deleteConnection(QName consInterface, QName provService, String provEndpoint)
            throws JBIException {
        this.log.start();
        if (consInterface == null) {
            throw new IllegalArgumentException("Consumer interface name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consInterface, provService,
                provEndpoint, this);
        this.registry.unregisterConnection(linkedEndpoint);
        this.log.end();
    }

    /**
     * @see EndpointRegistry#deleteConnection(QName, String, QName, String)
     */
    public void deleteConnection(QName consService, String consEndpoint, QName provService,
            String provEndpoint) throws JBIException {
        this.log.start();
        if (consService == null) {
            throw new IllegalArgumentException("Consumer service name must be non null");
        }
        if (consEndpoint == null) {
            throw new IllegalArgumentException("Consumer endpoint name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException("Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException("Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consService, consEndpoint, provService,
                provEndpoint, this);
        this.registry.unregisterConnection(linkedEndpoint);
        this.log.end();
    }

    /**
     * @see EndpointRegistry#deregisterExternalEndpoint(ServiceEndpoint)
     */
    public void deregisterExternalEndpoint(ServiceEndpoint endpoint) throws JBIException {
        this.log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        this.registry.unregisterExternalEndpoint(endpoint);
        this.log.end();
    }

    /**
     * @see EndpointRegistry#getEndpoint(QName, String) <ol>
     *  <li>Retreive the named endpoint</li>
     *  <li>If found, update the endpoint service of the found endpoint with
     *  this</li>
     *  </ol>
     */
    public ServiceEndpoint getEndpoint(QName service, String name) {
        this.log.call();
        return this.getEndpoint(service, name, true);
    }

    /**
     * @see EndpointRegistry#getEndpoint(QName, String,boolean) <ol>
     *  <li>Retrieve the named endpoint</li>
     *  <li>If found, update the endpoint service of the found endpoint with
     *  this</li>
     *  </ol>
     */
    public ServiceEndpoint getEndpoint(QName service, String name, boolean resolveLink) {
        this.log.start();
        if (service == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        if (name == null) {
            throw new IllegalArgumentException("Endpoint name must be non null");
        }
        AbstractEndpoint abstractEndpoint = null;
        try {
            abstractEndpoint = this.registry.getInternalEndpoint(service, name, resolveLink);
        } catch (RegistryException e) {
            // no endpoint found, return null;
        }

        this.log.end();
        return abstractEndpoint;
    }

    /**
     * @see EndpointRegistry#getEndpointDescriptorForEndpoint(ServiceEndpoint)
     */
    public Document getEndpointDescriptorForEndpoint(ServiceEndpoint endpoint) throws JBIException {
        this.log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        Document document = null;

        if ((endpoint instanceof AbstractEndpoint)) {
            document = ((AbstractEndpoint) endpoint).getDescription();
            if (document == null) {
                try {
                    // update the new endpoints
                    synchronized (this.containerName) {
                        this.updateNewEndpoints(false);
                    }

                    // retrieve the endpoint
                    AbstractEndpoint abstractEndpoint = this.registry.getInternalEndpoint(endpoint
                            .getServiceName(), endpoint.getEndpointName());

                    document = abstractEndpoint.getDescription();
                    // set both description and interfaces
                    ((AbstractEndpoint) endpoint).setDescription(document);
                    ((AbstractEndpoint) endpoint).setInterfaces(abstractEndpoint.getInterfaces());
                } catch (RegistryException e) {
                    this.log.error("Error while validating endpoint " + endpoint.getEndpointName()
                            + " for service " + endpoint.getServiceName(), e);
                }
            }
        }
        this.log.end();
        return document;
    }

    /**
     * @see EndpointRegistry#getExternalEndpointsForInterface(QName)
     */
    public ServiceEndpoint[] getExternalEndpointsForInterface(QName interfaceName) {
        this.log.start();

        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        try {
            endpoints = this.registry.getExternalEndpointsForInterface(interfaceName);
        } catch (RegistryException e) {
            this.log.error("Error while getting external endpoints for interface " + interfaceName,
                    e);
        }
        this.log.end();
        return endpoints;
    }

    /**
     * @see EndpointRegistry#getExternalEndpointsForService(QName)
     */
    public ServiceEndpoint[] getExternalEndpointsForService(QName serviceName) {
        this.log.start();

        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        ServiceEndpoint[] endpoints = null;
        endpoints = this.registry.getExternalEndpointsForService(serviceName);

        this.log.end();
        return endpoints;
    }

    /**
     * Return the interface for the specified endpoint
     * 
     * @param endpoint
     *      endpoint on which we want the interface names; must be non null
     * @return array of interfaces QName
     */
    public QName[] getInterfacesForEndpoint(AbstractEndpoint endpoint) {
        this.log.start();

        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }

        // update only the first time
        if (endpoint.getInterfaces() == null) {
            try {
                synchronized (this.containerName) {
                    this.updateNewEndpoints(false);
                }

                AbstractEndpoint abstractEndpoint = this.registry.getInternalEndpoint(endpoint
                        .getServiceName(), endpoint.getEndpointName());
                // set both description and interfaces
                endpoint.setInterfaces(abstractEndpoint.getInterfaces());
                endpoint.setDescription(abstractEndpoint.getDescription());
            } catch (Exception e) {
                this.log.error("Error while getting interfaces for endpoint "
                        + endpoint.getEndpointName() + " of service " + endpoint.getServiceName(),
                        e);
            }
        }

        this.log.end();
        return endpoint.getInterfaces();
    }

    /**
     * @see EndpointRegistry#getInternalEndpointsForInterface(QName)
     */
    public ServiceEndpoint[] getInternalEndpointsForInterface(QName interfaceName) {
        this.log.start();
        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        try {
            // Update the new endpoints
            synchronized (this.containerName) {
                this.updateNewEndpoints(false);
            }

            endpoints = this.registry.getInternalEndpointsForInterface(interfaceName);
        } catch (RegistryException e) {
            this.log.error("Error while getting internal endpoints for interface " + interfaceName,
                    e);
        } catch (JBIException e) {
            this.log.error("Error while getting internal endpoints for interface " + interfaceName,
                    e);
        }
        this.log.end();
        return endpoints;
    }

    /**
     * @see EndpointRegistry#getInternalEndpointsForService(QName)
     */
    public ServiceEndpoint[] getInternalEndpointsForService(QName serviceName) {
        this.log.start();

        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        endpoints = this.registry.getInternalEndpointsForService(serviceName);

        this.log.end();
        return endpoints;
    }

    /**
     * @see
     *  EndpointRegistry#isExchangeWithConsumerOkayForComponent(InternalEndpoint
     *  , MessageExchange)
     */
    public boolean isExchangeWithConsumerOkayForComponent(InternalEndpoint internalEndpoint,
            MessageExchange exchange) {
        this.log.start();
        ParameterCheckHelper.isNullParameterWithLog(internalEndpoint, "Endpoint", this.log);
        boolean test = true;
        try {
            // Retrieve a connection to the JMX server of the container were the
            // component is deployed
            RemoteJMXServer server = this.jmxService.createRemoteJMXServer(internalEndpoint
                    .getContainerName());
            // Ask the component if it is ok to accept the exchange
            Boolean bool = (Boolean) server.invoke(server.getAdminServiceMBeanName(),
                    RemoteJMXServer.IS_OK_WITH_CONS, new Object[] {
                            internalEndpoint.getComponentName(), internalEndpoint, exchange },
                    new String[] { String.class.getName(), ServiceEndpoint.class.getName(),
                            MessageExchange.class.getName() });
            test = bool.booleanValue();
        } catch (Exception e) {
            this.log.error("Problem in isExchangeWithConsumerOkayForComponent", e);
        }
        this.log.end();
        return test;
    }

    /**
     * @see
     *  EndpointRegistry#isExchangeWithProviderOkayForComponent(InternalEndpoint
     *  , MessageExchange)
     */
    public boolean isExchangeWithProviderOkayForComponent(InternalEndpoint internalEndpoint,
            MessageExchange exchange) {
        this.log.start();
        if (internalEndpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        boolean test = true;
        try {
            // Retrieve a connection to the JMX server of the container were the
            // component is deployed
            RemoteJMXServer server = this.jmxService.createRemoteJMXServer(internalEndpoint
                    .getContainerName());
            // Retrieve the endpoint description
            Boolean bool = (Boolean) server.invoke(server.getAdminServiceMBeanName(),
                    RemoteJMXServer.IS_OK_WITH_PROV, new Object[] {
                            internalEndpoint.getComponentName(), internalEndpoint, exchange },
                    new String[] { String.class.getName(), ServiceEndpoint.class.getName(),
                            MessageExchange.class.getName() });
            test = bool.booleanValue();
        } catch (Exception e) {
            this.log.error("Problem in isExchangeWithProviderOkayForComponent", e);
        }
        this.log.end();
        return test;
    }

    /**
     * @see EndpointRegistry#registerExternalEndpoint(ServiceEndpoint)
     */
    public void registerExternalEndpoint(ServiceEndpoint externalEndpoint) throws JBIException {
        this.log.start();
        if (externalEndpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        this.registry.registerExternalEndpoint(externalEndpoint);
        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws IllegalLifeCycleException {
        this.log = new LoggingUtil(this.logger);
        this.log.call();

        this.containerName = this.configurationService.getContainerConfiguration().getName();
        this.registry = new JNDIRegistry(this, this.configurationService, this.jndiService,
                this.log);
        try {
            this.registry.initContexts();
        } catch (RegistryException e) {
            throw new IllegalLifeCycleException(e.toString());
        }

    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws IllegalLifeCycleException {
        this.log.call();
        this.registry = null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#getDescription
     * (java.lang.String, java.lang.String)
     */
    public Document getDescription(String serviceName, String endpointName) throws JBIException {
        this.log.start();

        ServiceEndpoint serviceEndpoint = this
                .getEndpoint(QName.valueOf(serviceName), endpointName);

        this.log.end();
        return this.getEndpointDescriptorForEndpoint(serviceEndpoint);

    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#
     * getAllExternalEndpoints()
     */
    public List<Map<String, Object>> getAllExternalEndpoints() {
        this.log.start();

        List<Map<String, Object>> externalEndpointsMap = new ArrayList<Map<String, Object>>();
        ServiceEndpoint[] externalEndpoints = this.getExternalEndpointsForInterface(null);

        for (ServiceEndpoint serviceEndpoint : externalEndpoints) {
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) serviceEndpoint;
            Map<String, Object> endpointMap = new HashMap<String, Object>();
            endpointMap.put(KEY_COMPONENT_NAME, abstractEndpoint.getComponentName());
            endpointMap.put(KEY_CONTAINER_NAME, abstractEndpoint.getContainerName());
            endpointMap.put(KEY_ENDPOINT_NAME, abstractEndpoint.getEndpointName());
            endpointMap.put(KEY_SERVICE_NAME, abstractEndpoint.getServiceName());
            List<String> interfaceNames = this.interfaceQNameToString(abstractEndpoint);
            endpointMap.put(KEY_INTERFACE_NAMES, interfaceNames.toArray(new String[0]));
            endpointMap.put(KEY_TYPE, TYPE_EXTERNAL);
            externalEndpointsMap.add(endpointMap);
        }

        this.log.end();
        return externalEndpointsMap;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#
     * getAllInternalEndpoints()
     */
    public List<Map<String, Object>> getAllInternalEndpoints() {
        this.log.start();

        List<Map<String, Object>> internalEndpointsMap = new ArrayList<Map<String, Object>>();
        ServiceEndpoint[] internalEndpoints = this.getInternalEndpointsForInterface(null);

        for (ServiceEndpoint serviceEndpoint : internalEndpoints) {
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) serviceEndpoint;
            Map<String, Object> endpointMap = new HashMap<String, Object>();
            endpointMap.put(KEY_COMPONENT_NAME, abstractEndpoint.getComponentName());
            endpointMap.put(KEY_CONTAINER_NAME, abstractEndpoint.getContainerName());
            endpointMap.put(KEY_ENDPOINT_NAME, abstractEndpoint.getEndpointName());
            endpointMap.put(KEY_SERVICE_NAME, abstractEndpoint.getServiceName().toString());
            List<String> interfaceNames = this.interfaceQNameToString(abstractEndpoint);
            endpointMap.put(KEY_INTERFACE_NAMES, interfaceNames.toArray(new String[0]));
            if (serviceEndpoint instanceof LinkedEndpoint) {
                LinkedEndpoint linkEP = (LinkedEndpoint) serviceEndpoint;
                endpointMap.put(KEY_TYPE, linkEP.getToServiceName().toString() + "@"
                        + linkEP.getToEndpointName());
            } else {
                endpointMap.put(KEY_TYPE, TYPE_INTERNAL);
            }
            internalEndpointsMap.add(endpointMap);
        }

        this.log.end();
        return internalEndpointsMap;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#getAllEndpoints
     * ()
     */
    public List<Map<String, Object>> getAllEndpoints() {
        this.log.call();

        List<Map<String, Object>> allEndpointsMap = this.getAllExternalEndpoints();
        allEndpointsMap.addAll(this.getAllInternalEndpoints());

        return allEndpointsMap;
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.jbi.messaging.registry.EndpointRegistryMBean#
     * updateNewEndpoints(boolean)
     */
    public void updateNewEndpoints(boolean remoteContainerInvocation) throws JBIException {
        this.log.start();

        List<AbstractEndpoint> newLocalEndpoints = new ArrayList<AbstractEndpoint>();
        List<String> containersToUpdate = new ArrayList<String>();

        this.registry.retrieveAndDeleteLocalNewEndpoints(newLocalEndpoints, containersToUpdate);

        // Invoke update of remote containers only once
        if (!remoteContainerInvocation) {
            for (String containerToUpdate : containersToUpdate) {
                try {
                    RemoteJMXServer server = this.jmxService
                            .createRemoteJMXServer(containerToUpdate);

                    server.invoke(server.getEndpointServiceMBeanName(),
                            RemoteJMXServer.UPDATE_NEW_ENDPOINTS, new Object[] { true },
                            new String[] { boolean.class.getName() });

                } catch (Exception e) {
                    this.log.error("Failed to update new endpoints on the container '"
                            + containerToUpdate + "'", e);
                }
            }
        }

        // Update the local new endpoints
        for (AbstractEndpoint abstractEndpoint : newLocalEndpoints) {
            this.updateEndpointDescriptionAndInterfaces(abstractEndpoint);
            try {
                this.registry.validateEndpoint(abstractEndpoint);
            } catch (JBIException e) {
                this.log.error("Problem while updating new endpoint '"
                        + abstractEndpoint.getEndpointName() + "'", e);
            }
        }

        this.log.end();
    }

    /**
     * This method will update the given endpoint description and its
     * interfaces.
     * 
     * @param endpoint
     *      endpoint to update
     */
    private void updateEndpointDescriptionAndInterfaces(final AbstractEndpoint endpoint) {
        this.log.start();

        String description = null;
        List<QName> interfaces = new ArrayList<QName>();

        try {
            description = this.adminService.getServiceDescription(endpoint.getComponentName(),
                    endpoint);
        } catch (PetalsException e) {
            QName interfaze = new QName(endpoint.getServiceName().getNamespaceURI(),
                    UNRESOLVED_INTERFACE + endpoint.getServiceName().getLocalPart());
            this.log.warning("Failed to retrieve the description from the component '"
                    + endpoint.getComponentName() + "' for the endpoint '"
                    + endpoint.getEndpointName() + "'" + "'. Set its interface to " + interfaze, e);
            interfaces.add(interfaze);
        }

        try {
            // Get the whole description
            if (interfaces.size() == 0) {
                if (description == null) {
                    throw new PetalsException("The description retrieved for the endpoint '"
                            + endpoint.getEndpointName() + "' is null");
                } else {
                    // Get the interfaces
                    Document descriptionDocument = XMLUtil.createDocumentFromString(description);
                    if (descriptionDocument == null) {
                        throw new PetalsException("Failed to parse the description document '"
                                + description + "'");
                    }

                    endpoint.setDescription(descriptionDocument);
                    interfaces = WSDLUtil.getInterfacesForService(descriptionDocument, endpoint
                            .getServiceName());
                }
            }
        } catch (PetalsException e) {
            QName interfaze = new QName(endpoint.getServiceName().getNamespaceURI(),
                    UNRESOLVED_INTERFACE + endpoint.getServiceName().getLocalPart());
            this.log.warning("Failed to retrieve interfaces from the description of the endpoint '"
                    + endpoint.getEndpointName() + "'. Set its interface to " + interfaze, e);
            interfaces.add(interfaze);
        }

        endpoint.setInterfaces(interfaces.toArray(new QName[0]));

        this.log.end();
    }

    private List<String> interfaceQNameToString(AbstractEndpoint abstractEndpoint) {
        QName[] interfaces = abstractEndpoint.getInterfaces();
        List<String> interfaceNames = new ArrayList<String>();
        if (interfaces != null && interfaces.length > 0) {
            for (QName interfaceName : interfaces) {
                interfaceNames.add(interfaceName.toString());
            }
        } else {
            interfaceNames.add(UNDEFINED);
        }
        return interfaceNames;
    }

}
