/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PetalsAdminService.java 16:40:52 alouis $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.kernel.admin;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.Contingency;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.communication.topology.TopologyService;
import org.ow2.petals.jbi.management.deployment.DeploymentServiceMBean;
import org.ow2.petals.jbi.management.installation.InstallationServiceMBean;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.kernel.configuration.DomainConfiguration;
import org.ow2.petals.kernel.configuration.SubDomainConfiguration;
import org.ow2.petals.kernel.server.PetalsStopThread;
import org.ow2.petals.transport.Transporter;
import org.ow2.petals.util.LoggingUtil;

/**
 * This class is used for Petals global Administration
 * 
 * @author alouis, rnaudin - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.kernel.admin.PetalsAdminServiceInterface.class))
public class PetalsAdminServiceImpl implements PetalsAdminServiceInterface {

    /**
     * Logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * Deployment service fractal component
     */
    @Requires(name = "deployment", signature = org.ow2.petals.jbi.management.deployment.DeploymentServiceMBean.class)
    protected DeploymentServiceMBean deployment;

    /**
     * Installation service fractal component
     */
    @Requires(name = "installation", signature = org.ow2.petals.jbi.management.installation.InstallationServiceMBean.class)
    protected InstallationServiceMBean installation;

    /**
     * Transporter service fractal component
     */
    @Requires(name = "transporter", signature = org.ow2.petals.transport.Transporter.class)
    protected Transporter transporter;

    /**
     * Topology service fractal component
     */
    @Requires(name = "topology", signature = org.ow2.petals.communication.topology.TopologyService.class, contingency = Contingency.OPTIONAL)
    protected TopologyService topologyService;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * The container configuration
     */
    private ContainerConfiguration containerConfiguration;

    /**
     * The sub-domain configuration
     */
    private SubDomainConfiguration subDomainConfiguration;

    /**
     * The domain configuration
     */
    private DomainConfiguration domainConfiguration;

    /**
     * The Petals Stop thread instance, which stops the PEtALS container
     */
    private PetalsStopThread petalsStopThread;

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.admin.PetalsAdminServiceMBean#retrieveTopology()
     */
    public Set<Map<String, String>> retrieveTopology() throws PetalsException {
        this.log.start();

        Set<Map<String, String>> topologyMap = new HashSet<Map<String, String>>();

        topologyMap.add(this.getDomainMap());

        Set<SubDomainConfiguration> subdomainsConfiguration = this.getSubDomainsConfiguration();
        for (SubDomainConfiguration subdomain : subdomainsConfiguration) {
            topologyMap.add(this.getSubdomainMap(subdomain));
        }

        Set<ContainerConfiguration> containersConfiguration = this.getContainersConfiguration();
        for (ContainerConfiguration container : containersConfiguration) {
            Map<String, String> containerMap = this.getContainerMap(container);
            topologyMap.add(containerMap);
        }

        this.log.end();
        return topologyMap;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.admin.PetalsAdminServiceMBean#shutdownContainer()
     */
    public void shutdownContainer() throws Exception {
        this.log.start();

        this.log.info("PEtALS is shutting down...");

        if (this.domainConfiguration.getMode().equals(DomainConfiguration.DomainMode.DYNAMIC)) {
            this.log.info("  1/5 Undeploy Service Assemblies...");
        } else {
            this.log.info("  1/4 Undeploy Service Assemblies...");
        }

        try {
            this.deployment.shutdown();
        } catch (Exception e) {
            this.log.error("Can not shut down the Deployment service.", e);
        }

        if (DomainConfiguration.DomainMode.DYNAMIC.equals(this.domainConfiguration.getMode())) {
            this.log.info("  2/5 Uninstall Shared Libraries and Components...");
        } else {
            this.log.info("  2/4 Uninstall Shared Libraries and Components...");
        }

        try {
            this.installation.shutdown();
        } catch (Exception e) {
            this.log.error("Can not shut down the Installation service", e);
        }

        if (DomainConfiguration.DomainMode.DYNAMIC.equals(this.domainConfiguration.getMode())) {
            this.log.info("  3/5 Shutting down the Transporter...");
        } else {
            this.log.info("  3/4 Shutting down the Transporter...");
        }

        try {
            this.transporter.shutdown();
        } catch (Exception e) {
            this.log.error("Can not shut down the Transporter service", e);
        }

        if (DomainConfiguration.DomainMode.DYNAMIC.equals(this.domainConfiguration.getMode())) {
            this.log.info("  4/5 Remove the PEtALS node from the domain...");
            this.topologyService.unregisterContainer(this.containerConfiguration.getName());
        }

        if (DomainConfiguration.DomainMode.DYNAMIC.equals(this.domainConfiguration.getMode())) {
            this.log.info("  5/5 Stopping the server...");
        } else {
            this.log.info("  4/4 Stopping the server...");
        }

        // launch the stop of the Petals server in a different thread
        this.petalsStopThread.start();

        this.log.end();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.admin.PetalsAdminServiceMBean#stopContainer()
     */
    public void stopContainer() throws Exception {
        this.log.start();

        this.log.info("PEtALS is stopping...");
        // launch the stop of the Petals server in a different thread
        this.petalsStopThread.start();

        this.log.end();
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.kernel.admin.PetalsAdminServiceInterface#setPetalsStopThread(org.ow2.petals.kernel.server.PetalsStopThread)
     */
    public void setPetalsStopThread(PetalsStopThread petalsStopThread) {
        this.petalsStopThread = petalsStopThread;
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
     */
    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws IllegalLifeCycleException {
        this.log = new LoggingUtil(this.logger);
        this.log.call();

        this.containerConfiguration = this.configurationService.getContainerConfiguration();
        this.subDomainConfiguration = this.configurationService.getSubDomainConfiguration();
        this.domainConfiguration = this.configurationService.getDomainConfiguration();
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
     */
    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws IllegalLifeCycleException {
        this.log.call();
    }

    /**
     * Get the collection of all the container configurations of this domain.
     * 
     * @return Set of ContainerConfiguration
     * @throws PetalsException
     */
    private Set<ContainerConfiguration> getContainersConfiguration() throws PetalsException {
        this.log.start();

        Set<ContainerConfiguration> containers = null;

        if (this.topologyService == null) {
            // it is the standalone distribution
            containers = new HashSet<ContainerConfiguration>();
            containers.add(this.containerConfiguration);
        } else {
            containers = this.topologyService.getContainersConfiguration(null);
        }

        this.log.end();
        return containers;
    }

    /**
     * Get the collection of all the sub-domain configurations of this domain.
     * 
     * @return Set of Sub-domain
     * @throws PetalsException
     */
    private Set<SubDomainConfiguration> getSubDomainsConfiguration() throws PetalsException {
        this.log.start();

        Set<SubDomainConfiguration> subdomains = null;

        if (this.topologyService == null) {
            // it is the standalone distribution
            subdomains = new HashSet<SubDomainConfiguration>();
            subdomains.add(this.subDomainConfiguration);
        } else {
            subdomains = this.topologyService.getSubDomainsConfiguration();

        }

        this.log.end();
        return subdomains;
    }

    /**
     * Get the domain configuration in key/value form.
     * 
     * @return
     */
    private Map<String, String> getDomainMap() {
        log.start();

        Map<String, String> domainMap = new HashMap<String, String>();
        domainMap.put(TopologyKeyNames.TOPOLOGY_TYPE, DomainKeyNames.TOPOLOGY_TYPE);
        domainMap.put(DomainKeyNames.CONF_DOMAIN_NAME, this.domainConfiguration.getName());
        if (this.domainConfiguration.getDescription() != null) {
            domainMap.put(DomainKeyNames.CONF_DOMAIN_DESCRIPTION, this.domainConfiguration
                    .getDescription());
        }
        if (DomainConfiguration.DomainMode.DYNAMIC.equals(this.domainConfiguration.getMode())) {
            domainMap.put(DomainKeyNames.CONF_DOMAIN_MODE, "dynamic");
        } else if (DomainConfiguration.DomainMode.STATIC.equals(this.domainConfiguration.getMode())) {
            domainMap.put(DomainKeyNames.CONF_DOMAIN_MODE, "static");
        } else {
            domainMap.put(DomainKeyNames.CONF_DOMAIN_MODE, "standalone");
        }

        if (this.domainConfiguration.getNetworkMuticastIp() != null) {
            domainMap.put(DomainKeyNames.CONF_NETWORK_MULTICAST_IP, this.domainConfiguration
                    .getNetworkMuticastIp());
            domainMap.put(DomainKeyNames.CONF_NETWORK_MULTICAST_PORT, Integer
                    .toString(this.domainConfiguration.getNetworkMuticastPort()));
        }

        if (this.domainConfiguration.getJndiFactory() != null) {
            domainMap.put(DomainKeyNames.CONF_JNDI_FACTORY, this.domainConfiguration
                    .getJndiFactory());
        }
        if (this.domainConfiguration.getJndiHost() != null) {
            domainMap.put(DomainKeyNames.CONF_JNDI_HOST, this.domainConfiguration.getJndiHost());
        }
        if (this.domainConfiguration.getJndiPort() != 0) {
            domainMap.put(DomainKeyNames.CONF_JNDI_PORT, Integer.toString(this.domainConfiguration
                    .getJndiPort()));
        }

        log.end();
        return domainMap;
    }

    /**
     * Get the sub-domain configuration in key/value form.
     * 
     * @param subdomain
     * @return
     */
    private Map<String, String> getSubdomainMap(SubDomainConfiguration subdomain) {
        log.start();
        Map<String, String> subdomainMap = new Hashtable<String, String>();
        subdomainMap.put(TopologyKeyNames.TOPOLOGY_TYPE, SubDomainKeyNames.TOPOLOGY_TYPE);
        subdomainMap.put(SubDomainKeyNames.CONF_SUBDOMAIN_NAME, subdomain.getName());
        if (subdomain.getDescription() != null) {
            subdomainMap.put(SubDomainKeyNames.CONF_SUBDOMAIN_DESCRIPTION, subdomain
                    .getDescription());
        }
        subdomainMap.put(SubDomainKeyNames.CONF_NETWORK_TIME_SYNCHRONIZED, Boolean
                .toString(subdomain.isNetworkTimeSynchronized()));

        log.end();
        return subdomainMap;
    }

    /**
     * Get the container configuration in key/value form.
     * 
     * @param container
     * @return
     */
    private Map<String, String> getContainerMap(ContainerConfiguration container) {
        log.start();

        Map<String, String> containerMap = new Hashtable<String, String>();
        containerMap.put(TopologyKeyNames.TOPOLOGY_TYPE, ContainerKeyNames.TOPOLOGY_TYPE);
        containerMap.put(ContainerKeyNames.CONF_CONTAINER_NAME, container.getName());
        if (container.getDescription() != null) {
            containerMap.put(ContainerKeyNames.CONF_CONTAINER_DESCRIPTION, container
                    .getDescription());
        }
        containerMap.put(ContainerKeyNames.CONF_SUBDOMAIN_NAME, container.getSubdomainName());
        containerMap.put(ContainerKeyNames.CONF_HOST, container.getHost());

        if (container.getNetworkBindingPort() != 0) {
            containerMap.put(ContainerKeyNames.CONF_NETWORK_BIND_PORT, Integer.toString(container
                    .getNetworkBindingPort()));
        }

        containerMap.put(ContainerKeyNames.CONF_USER, container.getUser());
        containerMap.put(ContainerKeyNames.CONF_PWD, container.getPassword());

        containerMap.put(ContainerKeyNames.CONF_JMX_RMI_PORT, Integer.toString(container
                .getJmxRMIConnectorPort()));

        if (container.getJoramDomainPort() != 0) {
            containerMap.put(ContainerKeyNames.CONF_JORAM_DOMAIN_PORT, Integer.toString(container
                    .getJoramDomainPort()));
            containerMap.put(ContainerKeyNames.CONF_JORAM_TCP_PORT, Integer.toString(container
                    .getJoramTCPPort()));
        }
        if (container.getDreamTCPPort() != 0) {
            containerMap.put(ContainerKeyNames.CONF_DREAM_TCP_PORT, Integer.toString(container
                    .getDreamTCPPort()));
        }

        if (ContainerConfiguration.ContainerState.STARTED.equals(container.getState())) {
            containerMap.put(ContainerKeyNames.CONF_STATE, ContainerStates.STATE_STARTED);
        } else {
            containerMap.put(ContainerKeyNames.CONF_STATE, ContainerStates.STATE_STOPPED);
        }

        log.end();
        return containerMap;
    }
}
