/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: ClassLoaderServiceImpl.java 3585 2007-06-21 12:21:06Z garciap $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.platform.classloader;

import java.io.IOException;
import java.lang.reflect.Field;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.util.LoggingUtil;

/**
 * @version $Rev: 3585 $ $Date: 2007-06-21 14:21:06 +0200 (jeu, 21 jun 2007) $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>, Adrien
 *         LOUIS
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.platform.classloader.ClassLoaderService.class))
public class ClassLoaderServiceImpl implements ClassLoaderService {

    /**
     * the common loader name
     */
    public static final String COMMON_LOADER = "common";

    /**
     * the shared library class loader.
     */
    private SharedLibrariesClassLoader sharedLibrariesClassLoader;

    /**
     * The class loaders table.
     */
    private Map<String, PetalsClassLoader> classLoadersTable = new HashMap<String, PetalsClassLoader>();

    /**
     * Parent ClassLoader for all JBI ClassLoaders.
     */
    private ClassLoader parentClassLoader;

    /**
     * Default constructor.
     */
    public ClassLoaderServiceImpl() {
        super();
    }

    LoggingUtil log;

    /**
     * Logger
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * The Configuration service fractal component
     */
    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    // ----------------------------------------------------------------------
    // ClassLoadingService interface implementation
    // ----------------------------------------------------------------------

    /**
     * @see org.ow2.petals.platform.classloader.ClassLoaderService#containsClassLoader(java.lang.String)
     */
    public boolean containsClassLoader(String classLoaderId) {
        this.log.call();

        return this.classLoadersTable.containsKey(classLoaderId);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.platform.classloader.ClassLoaderService#createComponentClassLoader(java.lang.String,
     *      java.net.URL[], java.util.List, boolean, java.util.List)
     */
    public URLClassLoader createComponentClassLoader(String classLoaderId, URL[] baseUrls,
            List<String> classpathUrls, boolean useParentFirst, List<String> sharedLibrariesNameList)
            throws PetalsException {
        this.log.start();

        PetalsClassLoader cl = null;
        try {
            cl = new ComponentClassLoader(baseUrls, classpathUrls, this.sharedLibrariesClassLoader,
                    sharedLibrariesNameList);
            cl.setParentFirst(useParentFirst);
        } catch (IOException e) {
            throw new PetalsException("Error while creating a PEtALS class loader", e);
        }

        this.classLoadersTable.put(classLoaderId, cl);

        this.log.end();
        return cl;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.platform.classloader.ClassLoaderService#createSharedLibraryClassLoader(java.lang.String,
     *      java.net.URL[], java.util.List, boolean)
     */
    public URLClassLoader createSharedLibraryClassLoader(String classLoaderId, URL[] baseUrls,
            List<String> classpathUrls, boolean useParentFirst) throws PetalsException {
        this.log.start();

        SharedLibraryClassLoader cl = null;
        try {
            cl = new SharedLibraryClassLoader(baseUrls, classpathUrls, this.parentClassLoader);
            cl.setParentFirst(useParentFirst);
        } catch (IOException e) {
            throw new PetalsException("Error while creating a class loader", e);
        }

        this.classLoadersTable.put(classLoaderId, cl);
        this.sharedLibrariesClassLoader.addSharedLibraryClassLoader(classLoaderId, cl);

        this.log.end();
        return cl;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.platform.classloader.ClassLoaderService#deleteClassLoader(java.lang.String)
     */
    public void deleteClassLoader(String classLoaderId) {
        this.log.start();

        if (!this.containsClassLoader(classLoaderId)) {
            return;
        }

        PetalsClassLoader cl = this.classLoadersTable.remove(classLoaderId);
        if (cl instanceof SharedLibraryClassLoader) {
            this.sharedLibrariesClassLoader.removeSharedLibraryClassLoader(classLoaderId);
        }
        this.cleanCache(cl);

        this.log.end();
    }

    // //////////////////////////////////////////////////////////////
    // clear class loader caches
    // ////////////////////////////////////////////////////////////////

    /**
     * When uninstalling a component, some elements remains in the cache of
     * certain classes, and the classLoader can not release the JARS containing
     * those elements
     * 
     * referenced bug on Sun VM : 6232010
     * 
     */
    protected void cleanCache(PetalsClassLoader cLoader) {
        this.log.call();

        this.cleanCache(cLoader, "java.io.ObjectInputStream", "subclassAudits");
        this.cleanCache(cLoader, "java.io.ObjectOutputStream", "subclassAudits");
        this.cleanCache(cLoader, "java.io.ObjectStreamClass", "localDescs");
        this.cleanCache(cLoader, "java.io.ObjectStreamClass", "reflectors");
    }

    /**
     * Clear the cache contained in the specified class, for the specific field
     * (typically, streaming classes)
     * 
     * referenced bug on Sun VM : 6232010
     * 
     * @param the
     *            classLoader to clean (used when loading stream classes)
     * @param className
     *            the name of the class containing the cache field
     * @param fieldName
     *            the name of the cache field
     */
    @SuppressWarnings("unchecked")
    protected void cleanCache(PetalsClassLoader cLoader, String className, String fieldName) {

        Map cache = null;
        try {
            Class<?> clazz = cLoader.loadClass(className);
            for (Field field : clazz.getDeclaredFields()) {
                if (field.getType().equals(Map.class) || field.getType().equals(HashMap.class)) {
                    field.setAccessible(true);
                    cache = (Map) field.get(null);
                    if (cache != null) {
                        synchronized (cache) {
                            cache.clear();
                        }
                    }
                }
            }
        } catch (Throwable ex) {
            // class not loaded, nothing to do
        }
    }

    @LifeCycle(on = LifeCycleType.START)
    protected void start() {
        this.log = new LoggingUtil(this.logger);
        this.log.start();

        boolean isIsolated = this.configurationService.getContainerConfiguration()
                .isIsolateJBIClassLoaders();

        if (isIsolated) {
            this.parentClassLoader = ClassLoader.getSystemClassLoader();

        } else {
            this.parentClassLoader = ClassLoaderServiceImpl.class.getClassLoader();
        }
        this.classLoadersTable = new HashMap<String, PetalsClassLoader>();
        this.sharedLibrariesClassLoader = new SharedLibrariesClassLoader(this.parentClassLoader);

        this.log.end();
    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() {
        this.log.start();

        for (String classLoaderId : this.classLoadersTable.keySet()) {
            this.sharedLibrariesClassLoader.removeSharedLibraryClassLoader(classLoaderId);
            this.cleanCache(this.classLoadersTable.get(classLoaderId));
        }

        this.classLoadersTable = null;
        this.sharedLibrariesClassLoader = null;
        this.parentClassLoader = null;

        this.log.end();
    }
}
