/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PetalsClassLoader.java 7999 2008-04-30 12:49:39Z cdeneux $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.platform.classloader;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLStreamHandlerFactory;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.jar.JarFile;

import org.ow2.petals.platform.classloader.factory.URLFactory;
import org.ow2.petals.platform.classloader.locator.LocatorAbstract;
import org.ow2.petals.util.ConvertSpecialCharactersUtil;

/**
 * Petals class loader Do not implement loadClass(String) it causes a conflict
 * with the SharedLibrariesClassLoader
 * 
 * @version $Rev: 7999 $ $Date: 2008-04-30 14:49:39 +0200 (Wed, 30 Apr 2008) $
 * @since Petals 1.0
 * @author alouis, ddesjardins, Rafael Marins, Olivier Fabre
 */
public class PetalsClassLoader extends URLClassLoader {

    /**
     * Standard Java package binary name.
     */
    private static final String JAVA_PACKAGE = "java.";

    /**
     * Standard Java extensions package binary name.
     */
    private static final String JAVAX_PACKAGE = "javax.";

    /**
     * Bases URLs
     */
    private final URL[] bases;

    /**
     * Factories used to create URLs for URLClassLoader
     */
    private URLFactory[] factories;

    /**
     * List of known packages loaded from this class loader.
     */
    private Set<String> localKnownPackages;

    /**
     * Locators used to search jar/dir for files, directories, ...
     */
    private LocatorAbstract[] locators;

    /**
     * To use or not the parent first class loading delegation model.
     */
    private boolean parentFirst;

    /**
     * To use or not the package intelligence self-first mechanism.
     */
    private boolean usePackageIntelligence;

    /**
     * Create a class loader for petals
     * 
     * @param baseUrls
     *            installation root path in petals
     * @param classpathLocations
     *            jar defined in the classpath part of the jbi descriptor
     * @throws IOException
     */
    public PetalsClassLoader(URL[] baseUrls, List<String> classpathLocations) throws IOException {
        super(baseUrls, ClassLoader.getSystemClassLoader());
        this.bases = baseUrls;
        initialize(classpathLocations);
    }

    /**
     * Create a class loader for petals
     * 
     * @param baseUrls
     *            installation root path in petals
     * @param classpathLocations
     *            jar defined in the classpath part of the jbi descriptor
     * @param parent
     * @throws IOException
     */
    public PetalsClassLoader(URL[] baseUrls, List<String> classpathLocations, ClassLoader parent)
            throws IOException {
        super(baseUrls, parent);
        this.bases = baseUrls;
        initialize(classpathLocations);
    }

    /**
     * Create a class loader for petals
     * 
     * @param baseUrls
     *            installation root path in petals
     * @param classpathLocations
     *            jar defined in the classpath part of the jbi descriptor
     * @param parent
     *            parent class loader
     * @param factory
     *            url factory
     * @throws IOException
     */
    public PetalsClassLoader(URL[] baseUrls, List<String> classpathLocations, ClassLoader parent,
            URLStreamHandlerFactory factory) throws IOException {
        super(baseUrls, parent, factory);
        this.bases = baseUrls;
        initialize(classpathLocations);
    }

    // ----------------------------------------------------------------------
    // Getters and Setters
    // ----------------------------------------------------------------------

    public void close() throws Throwable {
        finalize();
    }

    /**
     * @return Returns the bases.
     */
    public URL[] getBases() {
        return bases;
    }

    /**
     * @return Returns a String representation of the classpath used by this
     *         classloader
     */
    public String getClasspath() {
        URL[] urls = getURLs();
        StringBuffer cp = new StringBuffer();
        for (int i = 0; i < urls.length; i++) {
            String url = urls[i].getFile();
            // do not add URL with !/ inside
            if (url.indexOf("!/") == -1) {
                cp.append(File.pathSeparator + url);
            }
        }
        return cp.toString();
    }

    /**
     * Return a URL of the resource
     * 
     * @param resourceName
     * @return url of the resource (null if the resource is not found)
     */
    @Override
    public URL getResource(String resourceName) {
        URL resource = null;
        // parent first
        if (useParentFirst(resourceName, false)) {
            resource = this.getParent().getResource(resourceName);
            if (resource == null) {
                resource = findResource(resourceName);
            }
        } else {
            // self first
            resource = findResource(resourceName);
            if (resource == null) {
                resource = this.getParent().getResource(resourceName);
            }
        }
        return resource;
    }

    /**
     * Get a stream from a resource in the classpath
     * 
     * @param resourceName
     *            name of the resource
     * @return stream of the resource (null if the resource is not found)
     */
    @Override
    public InputStream getResourceAsStream(String resourceName) {
        InputStream inputStream = null;
        // parent first
        if (useParentFirst(resourceName, false)) {
            inputStream = this.getParent().getResourceAsStream(resourceName);
            if (inputStream == null) {
                URL url = findResource(resourceName);
                inputStream = getStream(url);
            }
        } else {
            // self first
            URL url = findResource(resourceName);
            if (url == null) {
                url = this.getParent().getResource(resourceName);
            }
            inputStream = getStream(url);
        }
        return inputStream;
    }

    /**
     * Return a enumeration of resources
     * 
     * @see java.lang.ClassLoader#getResources(java.lang.String)
     */
    @Override
    public Enumeration<URL> getResources(String resourceName) throws IOException {
        Enumeration<URL> resources = null;
        // parent first
        if (useParentFirst(resourceName, false)) {
            resources = this.getParent().getResources(resourceName);
            if (resources == null || !resources.hasMoreElements()) {
                resources = findResources(resourceName);
            }
        } else {
            // self first
            resources = findResources(resourceName);
            if (resources == null || !resources.hasMoreElements()) {
                resources = this.getParent().getResources(resourceName);
            }
        }
        return resources;
    }

    /**
     * @return Returns the useParentFirst.
     */
    public boolean isParentFirst() {
        return parentFirst;
    }

    // ----------------------------------------------------------------------
    // overriding methods
    // ----------------------------------------------------------------------

    /**
     * @return Returns the usePackageIntelligence.
     */
    public boolean isUsingPackageIntelligence() {
        localKnownPackages = null;
        return usePackageIntelligence;
    }

    /**
     * @param use
     *            The useParentFirst to set.
     */
    public void setParentFirst(boolean use) {
        this.parentFirst = use;
    }

    /**
     * @param use
     *            The usePackageIntelligence to set.
     */
    public void setUsingPackageIntelligence(boolean use) {
        localKnownPackages = new HashSet<String>();
        this.usePackageIntelligence = use;
    }

    /**
     * @return Returns a String representation of the PetalsClassLoader
     */
    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();

        sb.append("classloader : " + getClass().getName() + "\n");
        sb.append("\tmodules bases (not in loader!) : \n");
        for (int i = 0; i < bases.length; i++) {
            sb.append("\t\t -" + bases[i] + "\n");
        }
        sb.append("\trepositories :\n");
        URL[] rep = getURLs();
        for (int i = 0; i < rep.length; i++) {
            sb.append("\t\t -" + rep[i] + "\n");
        }
        sb.append("\tparent : " + getParent() + "\n");

        return sb.toString();
    }

    /**
     * Add specified location into the repository. If location is found in
     * multiple URLs of the bases, new URLs will be added multiple times.
     * 
     * @param location
     *            an entry name (for a jar) or a path name (for a directory)
     * @throws IOException
     *             when constructed URL is malformed
     */
    protected final void addInRepository(String location) throws IOException {
        // test existence in each base URL
        for (int i = 0; i < bases.length; i++) {
            if (locators[i].hasDirectory(location)) {
                if (location.charAt(location.length() - 1) != '/') {
                    location = location + "/";
                }
                addURL(factories[i].getURL(location));
            } else if (locators[i].hasFile(location)) {
                addURL(factories[i].getURL(location));
            }
        }
    }

    // ----------------------------------------------------------------------
    // implementation specific methods
    // ----------------------------------------------------------------------

    /**
     * Add a package name to the known package
     * 
     * @param packageName
     *            package name
     */
    protected void addLocalKnownPackage(String packageName) {
        if (localKnownPackages != null) {
            localKnownPackages.add(packageName);
        }
    }

    /**
     * Return the package name of a class
     * 
     * @param className
     *            complete name of a class
     * @return package name
     */
    protected String getPackageName(String className) {
        int lastSeparator = className.lastIndexOf('.');
        return className.substring(0, lastSeparator);
    }

    /**
     * Return an input stream from a url
     * 
     * @param url
     * @return
     */
    protected InputStream getStream(URL url) {
        InputStream inputStream = null;
        if (url != null) {
            try {
                if ("file".equals(url.getProtocol())) {
                    // The resource is a file
                    inputStream = new FileInputStream(url.getPath());
                } else if ("jar".equals(url.getProtocol())) {
                    // The resource is contain into a Jar
                    try {
                        String fileName = url.getPath().substring(0, url.getPath().indexOf("!"));
                        fileName = fileName.substring(5);
                        JarFile jarFile = new JarFile(fileName);
                        String entryName = url.getPath().substring(url.getPath().indexOf("!") + 2);
                        // Change spaces
                        entryName = ConvertSpecialCharactersUtil.convertPath(entryName);
                        inputStream = jarFile.getInputStream(jarFile.getEntry(entryName));
                    } catch (Exception e) {
                        inputStream = null;
                    }
                }
            } catch (FileNotFoundException e) {
                inputStream = null;
            }
        }
        return inputStream;
    }

    /**
     * Initialize the petals class loader
     * 
     * @param classpathLocations
     *            elements of the classpath
     */
    private final void initialize(final List<String> classpathLocations) throws IOException {
        this.parentFirst = true;
        this.usePackageIntelligence = false;
        factories = new URLFactory[bases.length];
        locators = new LocatorAbstract[bases.length];
        // Create factories and locator for each base URL
        for (int i = 0; i < bases.length; i++) {
            factories[i] = URLFactory.getFactory(bases[i]);
            locators[i] = LocatorAbstract.getLocator(bases[i]);
        }
        // Add all classpathLocations into the repository
        for (String location : classpathLocations) {
            addInRepository(location);
        }
    }

    /**
     * Check if a package is in the known package list
     * 
     * @param packageName
     *            package name
     * @return true if the package is already known
     */
    protected boolean knowLocalPackage(String packageName) {
        boolean known = false;
        if (localKnownPackages != null) {
            known = localKnownPackages.contains(packageName);
        }
        return known;
    }

    /**
     * Load a class Use the parent classloader if parentFirst is set to true
     * else use its own url classloader to load the class
     * 
     * @param className
     *            name of the class to load
     * @param resolve
     *            boolean to resolve or not the class
     */
    @Override
    protected synchronized Class<?> loadClass(String className, boolean resolve)
            throws ClassNotFoundException {
        Class<?> clazz = null;
        clazz = findLoadedClass(className);
        if (clazz == null) {
            // parent first
            if (useParentFirst(className, true)) {
                try {
                    clazz = this.getParent().loadClass(className);
                } catch (Throwable e) {
                    clazz = findClass(className);
                }
            } else {
                // self first
                try {
                    clazz = findClass(className);
                } catch (Throwable e) {
                    // The class may be already loaded from a parent class
                    // loader
                    clazz = this.getParent().loadClass(className);
                }
            }
        }

        if (clazz == null) {
            throw new ClassNotFoundException(className);
        }

        if (resolve) {
            resolveClass(clazz);
        }
        return clazz;
    }

    /**
     * Method to know if we choose to use the parent class loader first
     * 
     * @param resourceName
     *            resource name
     * @param usingForClass
     *            resource as a class
     * @return boolean
     */
    protected boolean useParentFirst(String resourceName, boolean usingForClass) {
        boolean useParentFirst = isParentFirst();

        // try self-first based on intelligent mechanism of known packages
        if (useParentFirst) {
            if (isUsingPackageIntelligence() && usingForClass) {
                String packageName = getPackageName(resourceName);
                useParentFirst = !knowLocalPackage(packageName);
            }
        } else {
            // always use parent class loader when loading java packages
            useParentFirst = resourceName.startsWith(JAVA_PACKAGE)
                    || resourceName.startsWith(JAVAX_PACKAGE);
        }
        return useParentFirst;
    }
}
