/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.transport.util;

import java.io.File;
import java.io.IOException;
import java.net.ConnectException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.jms.ConnectionFactory;

import org.apache.commons.io.FileUtils;
import org.objectweb.joram.client.jms.Destination;
import org.objectweb.joram.client.jms.Queue;
import org.objectweb.joram.client.jms.Topic;
import org.objectweb.joram.client.jms.admin.AdminException;
import org.objectweb.joram.client.jms.admin.AdminModule;
import org.objectweb.joram.client.jms.local.LocalConnectionFactory;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.ow2.petals.kernel.api.server.PetalsException;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.transport.TransportException;
import org.ow2.petals.transport.platform.joram.JoramConnection;

import fr.dyade.aaa.agent.AgentServer;

/**
 * Utility class to control the JOrAM agent server.
 * 
 * @author Christophe HAMERLING - eBMWebSourcing
 * @author Roland NAUDIN - eBMWebSourcing
 * 
 */
public class JoramControl {

    /**
     * Instance of the Joram control class
     */
    private static JoramControl instance;

    /**
     * The local connection of this Joram agent server
     */
    private ConnectionFactory connectionFactory;

    /**
     * Singleton access
     * 
     * @return
     */
    public static synchronized JoramControl getInstance() {
        if (instance == null) {
            instance = new JoramControl();
        }
        return instance;
    }

    /**
     * Creates a new instance of {@link JoramControl}
     * 
     */
    private JoramControl() {
    }

    /**
     * Clean the workspace
     * 
     */
    public final void cleanAgentServerWorkspace(File joramDir, boolean forceDelete) {
        for (File file : joramDir.listFiles()) {
            if (file.isDirectory()) {
                // We are in a Joram backup directory
                boolean isLock = false;
                for (File joramFile : file.listFiles()) {
                    if ("lock".equals(joramFile.getName())) {
                        // We are looking for a lock file
                        isLock = true;
                        break;
                    }
                }
                if (isLock || forceDelete) {
                    // This Joram server was not shutdown correctly
                    // We need to remove the backup directory
                    try {
                        FileUtils.forceDelete(file);
                    } catch (IOException e) {
                        // Do nothing
                    }
                }
            }
        }
    }

    /**
     * Initialize the agent server
     * 
     * @param agentDir
     * @param id
     * @param factory
     * @throws Exception
     */
    private void doInitAgentServer(File agentDir, short id) throws Exception {
        System.setProperty(AgentServer.CFG_DIR_PROPERTY, agentDir.getAbsolutePath());

        AgentServer.init(id, agentDir.getAbsolutePath() + File.separator + id,
                getLoggerFactoryInstance());
    }

    /**
     * Test if the agent server is started
     * 
     * @return
     */
    public final boolean isAgentServerRunning() {
        return 4 == AgentServer.getStatus();
    }

    /**
     * Start the agent server.
     * 
     * @throws Exception
     *             the exception throwed by the joram agent
     */
    public final void startAgentServer() throws Exception {
        AgentServer.start();
    }

    /**
     * Stop the agent server
     * 
     * @throws AdminException
     * @throws UnknownHostException
     * @throws NumberFormatException
     * @throws ConnectException
     * 
     */
    public final void stopAgentServer()
            throws ConnectException, NumberFormatException, UnknownHostException, AdminException {
        AgentServer.stop();
    }

    /**
     * Retrieves the Joram queue sizes managed by the given container
     * 
     * @param containerConfiguration
     *            the container configuration (name, Joram login and
     *            password...)
     * @return a map containing queue sizes ordered by component
     * @throws TransportException
     *             if an error occurs during the connection and dialog with the
     *             Joram AdminModule
     */
    @SuppressWarnings("unchecked")
    public final HashMap<String, Integer> getQueueSizes(
            ContainerConfiguration containerConfiguration) throws TransportException {
        HashMap<String, Integer> componentsQueues = new HashMap<String, Integer>();
        try {
            List<Destination> destinations = new ArrayList<Destination>();
            destinations = AdminModule.getDestinations();
            for (Destination destination : destinations) {
                String componentName = JoramConnection.getComponentName(destination.getAdminName(),
                        containerConfiguration.getName());
                if (destination.isQueue() && componentName != null) {
                    Queue componentQueue = (Queue) destination;
                    componentsQueues.put(componentName, componentQueue.getPendingMessages());
                }
            }
        } catch (ConnectException e) {
            throw new TransportException("Can't retrieve Joram destinations (queues and topics)", e);
        } catch (AdminException e) {
            throw new TransportException("Can't retrieve Joram destinations (queues and topics)", e);
        } catch (NumberFormatException e) {
            throw new TransportException("Can't connect to admin module", e);
        }
        return componentsQueues;
    }

    /**
     * Retrieves the Joram component max queue size. For the moment, all
     * components have the same max queue size
     * 
     * @param containerConfiguration
     *            the container configuration (name, Joram login and
     *            password...)
     * @return the component max queue size
     * @throws TransportException
     *             if an error occurs during the connection and dialog with the
     *             Joram AdminModule
     */
    @SuppressWarnings("unchecked")
    public final int getQueueMaxSize(ContainerConfiguration containerConfiguration)
            throws TransportException {
        int result = 0;
        try {
            List<Destination> destinations = new ArrayList<Destination>();
            destinations = AdminModule.getDestinations();

            for (Destination destination : destinations) {
                String componentName = JoramConnection.getComponentName(destination.getAdminName(),
                        containerConfiguration.getName());
                if (destination.isQueue() && componentName != null) {
                    Queue componentQueue = (Queue) destination;
                    result = componentQueue.getNbMaxMsg();
                    break;
                }
            }
        } catch (ConnectException e) {
            throw new TransportException("Can't retrieve Joram destinations (queues and topics)", e);
        } catch (AdminException e) {
            throw new TransportException("Can't retrieve Joram destinations (queues and topics)", e);
        } catch (NumberFormatException e) {
            throw new TransportException("Can't connect to admin module", e);
        }
        return result;
    }

    /**
     * Add a Joram server to the Joram platform
     * 
     * @param id
     * @param host
     * @param domain
     * @param domainPort
     * @param name
     * @param tcpPort
     * @throws ConnectException
     * @throws AdminException
     */
    public void addJoramServer(int id, String host, String domain, int domainPort, String name,
            int tcpPort) throws ConnectException, AdminException {
        AdminModule.addServer(id, host, domain, domainPort, name,
                new String[] { "org.objectweb.joram.mom.proxies.tcp.TcpProxyService" },
                new String[] { String.valueOf(tcpPort) });
    }

    /**
     * 
     * @param containerInformation
     * @param joramDir
     * @throws PetalsException
     */
    public void initAgentServer(File joramDir, short containerId) throws PetalsException {
        if (!isAgentServerRunning()) {
            try {
                cleanAgentServerWorkspace(joramDir, false);
                doInitAgentServer(joramDir, containerId);
            } catch (Exception e) {
                fr.dyade.aaa.agent.AgentServer.stop(true);
                throw new PetalsException("Joram Agent can not be initialized", e);
            }
        }
    }

    /**
     * Remove a container from JOrAM agent. Connects to the JOrAM module, remove
     * the server and disconnect from the admin module.
     * 
     * @param containerConfiguration
     * @throws PetalsException
     */
    public void removeContainer(ContainerConfiguration containerConfiguration)
            throws PetalsException {
        if (containerConfiguration == null) {
            throw new IllegalArgumentException("Container configuration cannot be null");
        }

        try {
            AdminModule.removeServer(Integer.parseInt(containerConfiguration.getName()));
        } catch (ConnectException e) {
            throw new PetalsException("Can not remove container from JORAM agent", e);
        } catch (AdminException e) {
            throw new PetalsException("Can not remove container from JORAM agent", e);
        }
    }

    /**
     * Remove the local server.
     * 
     * @param containerConfiguration
     * @throws PetalsException
     */
    public void removeLocalContainer(ContainerConfiguration containerConfiguration)
            throws PetalsException {
        if (containerConfiguration == null) {
            throw new IllegalArgumentException("Container configuration cannot be null");
        }

        try {
            AdminModule.removeServer(Integer.parseInt(containerConfiguration.getName()));

        } catch (ConnectException e) {
            throw new PetalsException("Can not remove container from JORAM agent", e);
        } catch (AdminException e) {
            throw new PetalsException("Can not remove container from JORAM agent", e);
        }
    }

    /**
     * Get a local connection factory
     * 
     */
    public ConnectionFactory getConnectionFactory() {
        if (connectionFactory == null) {
            connectionFactory = LocalConnectionFactory.create();

            // FIXME optim multithread
            // connectionFactory.getParameters().multiThreadSync = true;
            // connectionFactory.getParameters().multiThreadSyncDelay = 1;
        }
        return connectionFactory;
    }

    /**
     * Connect the Admin module to the collocated server
     * 
     * @param containerInformation
     * @throws AdminException
     * @throws UnknownHostException
     * @throws NumberFormatException
     * @throws ConnectException
     * @throws Exception
     * 
     * @see AdminModule#collocatedConnect(String, int, String, String, int,
     *      String)
     */
    public void connectToAdminModule(ContainerConfiguration containerConfiguration)
            throws ConnectException, NumberFormatException, UnknownHostException, AdminException {
        if (containerConfiguration == null) {
            throw new IllegalArgumentException("Container configuration can not be null");
        }

        AdminModule.collocatedConnect(containerConfiguration.getUser(), containerConfiguration
                .getPassword());
    }

    /**
     * Disconnect from the JORAM admin module
     * 
     */
    public void disconnectFromAdminModule() {
        AdminModule.disconnect();
    }

    /**
     * Create a new topic
     * 
     * @param topicName
     *            the name of the topic to be created
     * @see Topic#create()
     */
    public synchronized Topic createTopic(String topicName) throws ConnectException, AdminException {
        Topic topic = Topic.create(topicName);
        topic.setFreeReading();
        topic.setFreeWriting();
        return topic;
    }

    /**
     * Create a new queue
     * 
     * @param queueName
     *            the name of the queue to be created
     * @see Topic#create()
     */
    public synchronized Queue createQueue(String queueName) throws ConnectException, AdminException {
        Queue queue = Queue.create(queueName);
        queue.setFreeReading();
        queue.setFreeWriting();
        return queue;
    }

    /**
     * Check if the given destination exists.
     * 
     * @param destination
     *            the destination to check
     */
    @SuppressWarnings("unchecked")
    public synchronized boolean isDestinationExist(Destination destination, int serverId)
            throws ConnectException, AdminException {
        List serverDestinations = AdminModule.getDestinations(serverId);

        for (Object serverDestination : serverDestinations) {
            if (((Destination) serverDestination).getName().equals(destination.getName())) {
                return true;
            }
        }

        return false;
    }

    /**
     * Get the logger factory to be used by JORAM. It must have been properly
     * initialized at the Petals server start.
     * 
     * @return
     */
    private LoggerFactory getLoggerFactoryInstance() {
        return Monolog.getMonologFactory();
    }
}
