/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.webconsole;

import java.io.File;

import org.mortbay.jetty.Connector;
import org.mortbay.jetty.Server;
import org.mortbay.jetty.nio.SelectChannelConnector;
import org.mortbay.jetty.webapp.WebAppContext;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.kernel.configuration.ConfigurationService;
import org.ow2.petals.kernel.configuration.ContainerConfiguration;
import org.ow2.petals.util.LoggingUtil;

/**
 * 
 * Created on 29 janv. 08
 * 
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since PEtALS 2.1
 * 
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.ow2.petals.webconsole.WebConsoleInterface.class))
public class WebConsoleImpl implements WebConsoleInterface {

    private static final String PROPERTY_WEBCONSOLE_PORT = "user.webconsole.port";

    private static final int DEFAULT_PORT = 7878;

    /**
     * Monolog logger
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
     * Logging util
     */
    private LoggingUtil log;

    /**
     * The port used by Jetty server
     */
    private int port = DEFAULT_PORT;

    /**
     * The host name used by jetty server
     */
    private String host;

    @Requires(name = "configuration", signature = org.ow2.petals.kernel.configuration.ConfigurationService.class)
    private ConfigurationService configurationService;

    /**
     * The jetty server
     */
    private Server server;

    @LifeCycle(on = LifeCycleType.START)
    protected void start() throws IllegalLifeCycleException {
        this.log = new LoggingUtil(this.logger);

        final ContainerConfiguration containerConfiguration = this.configurationService
                .getContainerConfiguration();

        // get the http port to be used
        if (containerConfiguration.getUserConfiguration() != null) {
            if (containerConfiguration.getUserConfiguration().containsKey(PROPERTY_WEBCONSOLE_PORT)) {
                try {
                    this.port = Integer.parseInt(containerConfiguration.getUserConfiguration().get(
                            PROPERTY_WEBCONSOLE_PORT));
                } catch (final NumberFormatException e) {
                    this.log.error("The parameter " + PROPERTY_WEBCONSOLE_PORT
                            + " is not set properly", e);
                    throw e;
                }
            }
        }
        System.setProperty("webconsole.port", Integer.toString(this.port));

        // get the host
        this.host = containerConfiguration.getHost();
        System.setProperty("webconsole.host", this.host);

        // create embedded jetty and webconsole application
        this.server = new Server();
        final Connector connector = new SelectChannelConnector();

        connector.setHost(this.host);
        connector.setPort(this.port);
        this.server.setConnectors(new Connector[] { connector });

        final WebAppContext webapp = new WebAppContext();
        webapp.setContextPath("/");
        webapp.setWar(containerConfiguration.getInstallDirectoryPath()
                + "/webapps/petals-webconsole.war");
        final File workPath = new File(containerConfiguration.getInstallDirectoryPath(), "work");
        webapp.setTempDirectory(new File(workPath, "petals-webconsole"));
        webapp.setExtractWAR(true);

        this.server.setHandler(webapp);

        // start embedded Jetty
        try {
            this.log.debug("Starting embedded Jetty...");
            this.server.start();
            this.log.info("PEtALS Web console is available at 'http://" + this.host + ":"
                    + this.port + "'");
        } catch (final InterruptedException e) {
            this.log.warning("Jetty server interrupted !", e);
        } catch (final Exception e) {
            this.log.warning("Can not start Jetty server !", e);
        }
    }

    /*
     * 
     */
    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws IllegalLifeCycleException {
        this.log.info("Stopping embedded Jetty...");
        try {
            this.server.stop();
        } catch (final Exception e) {
            this.log.warning("Problem while stopping Jetty server", e);
        }
    }
}
